/*
	buffer.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	Provides buffered read/write for file descriptors.

	This file uses 4 character tabstops.

	Version 1.00	02/27/1998
	Version 1.01	05/18/1999		In bffree(), free() the buffer too.
	Version 1.02	03/29/2000		Split off code from bfread() and bfwrite() 
	                                into bf[rw]bytes().
	Version 1.03	12/21/2001		Added bfinit(), bfamount(), bfsize(), 
	                                bfflush(), bfclear(), bfaccount4read(),
	                                bfstrcat(), bfaddch(), bfstrstr(), 
	                                bfstrchr()
	Version 1.04	01/04/2002		Don't wrap the readp and the writep when 
	                                they reach the end of the buffer.  Instead,
	                                the buffer will be re-initialized when we've
	                                emptied it.
	Version 1.05	01/04/2002		Use 'unsigned char' everywhere.  Added 
	                                bfstrncat().  Minor changes to allow a BUFFER
	                                to contain embedded nulls.
	Version 1.06	01/25/2002		Added bfpeek()
	Version 1.07	02/27/2002		Don't print the "buffer is empty" message in 
	                                bfwbytes(), as this is a common case.  Use 
	                                DBG_VINF in bfdump(), instead of DBG_INF.
*/

#include "telnetcpcd.h"

/* internal functions */
void bfinit(struct buffer_t *buff);
void bfaccount4read(struct buffer_t *buff,int nbytes);
void bfaccount4write(struct buffer_t *buff,int nbytes);
int bfrbytes(struct buffer_t *buff);
int bfwbytes(struct buffer_t *buff);

struct buffer_t *bfmalloc(char *label,int size)
{
	struct buffer_t *buff;

	buff = calloc(1,sizeof(struct buffer_t));
	if (buff == NULL) {
		debug_perror("bfmalloc()");
		return(NULL);
	}
	if (size > 0) {
		buff->size = size;
	} else {
		buff->size = BUFSIZ;
	}
	buff->buffp = calloc(1,(buff->size + 1));
	if (buff->buffp == NULL) {
		debug_perror("bfmalloc()");
		free(buff);
		return(NULL);
	}
	if (label != NULL) {
		buff->label = (unsigned char *) strdup(label);
	} else {
		buff->label = (unsigned char *) strdup("");
	}
	if (buff->label == NULL) {
		debug_perror("bfmalloc()");
		free(buff->buffp);
		free(buff);
		return(NULL);
	}
	buff->eof = 0;
	bfinit(buff);
	return(buff);
}

void bfinit(struct buffer_t *buff)
{
	if (buff == NULL) return;

	buff->readp = buff->buffp;
	buff->writep = buff->buffp;
	buff->tailp = buff->buffp + buff->size;
	buff->nbuffered = 0;
	memset(buff->buffp,(int) '\0',buff->size);
}

void bfaccount4read(struct buffer_t *buff,int nbytes)
{
	if (buff == NULL) return;

	buff->nbuffered += nbytes;
	buff->readp += nbytes;
}

void bfaccount4write(struct buffer_t *buff,int nbytes)
{
	if (buff == NULL) return;

	buff->nbuffered -= nbytes;
	if (buff->nbuffered > 0) {
		buff->writep += nbytes;
	} else {
		bfinit(buff);
	}
}

void bffree(struct buffer_t *buff)
{
	if (buff == NULL) return;

	if (buff->label != NULL)
		free(buff->label);
	if (buff->buffp != NULL)
		free(buff->buffp);
	free(buff);
}

int bfread(int fd,struct buffer_t *buff)
{
	extern int errno;
	int bytes;
	int n;

	if (buff == NULL) return(0);

	bytes = bfrbytes(buff);
	if (bytes <= 0) 
		return(bytes);

	n = read(fd,buff->readp,bytes);
	if (n < 0) {
		debug_perror("bfread()");
		if ((errno != EINTR) && (errno != EAGAIN)) {
			return(n);
		} else {
			return(0);
		}
	} else if (n == 0) {
		debug(DBG_INF,"bfread(): eof on fd %d",fd);
		buff->eof = 1;
		return(0);
	}
	/* else (n > 0) */
	debug(DBG_VINF,"bfread(): read(%d,readp,%d) = %d",fd,bytes,n);
	bfaccount4read(buff,n);
	debug(DBG_VINF,"bfread(): nbuffered=%d",buff->nbuffered);
	return(n);
}

/*
	calculate the number of bytes we could read into the buffer
*/
int bfrbytes(struct buffer_t *buff)
{
	int bytes;

	if (buff == NULL) return(0);

	if (buff->readp > buff->writep) {
		bytes = buff->tailp - buff->readp;
	} else if (buff->readp < buff->writep) {
		bytes = buff->writep - buff->readp;
	} else {
		/* (buff->readp == buff->writep) */
		if (buff->nbuffered == 0) {			/* buffer is empty */
			bytes = buff->size;
		} else {							/* buffer is full */
			bytes = 0;
		}
	}
	if (bytes <= 0) {
		if (bytes == 0) {
			debug(DBG_INF,"bfread() buffer \"%s\" is full",buff->label);
		} else {
			debug(DBG_ERR,"bfread() buffer \"%s\" is corrupt: bytes=%d",buff->label,bytes);
		}
	}
	return(bytes);
}

int bfwrite(int fd,struct buffer_t *buff)
{
	extern int errno;
	int bytes;
	int n;

	if (buff == NULL) return(0);

	bytes = bfwbytes(buff);
	if (bytes <= 0) 
		return(bytes);

	n = write(fd,buff->writep,bytes);
	if (n < 0) {
		debug_perror("bfwrite()");
		if ((errno != EINTR) && (errno != EAGAIN)) {
			return(n);
		} else {
			return(0);
		}
	} else if (n == 0) {
		debug(DBG_INF,"bfwrite(): write() returned 0");
		return(0);
	}
	/* else (n > 0) */
	debug(DBG_VINF,"bfwrite(): write(%d,writep,%d) = %d",fd,bytes,n);
	bfaccount4write(buff,n);
	debug(DBG_VINF,"bfwrite(): nbuffered=%d",buff->nbuffered);
	return(n);
}

/*
	calculate the number of bytes we could write from the buffer
*/
int bfwbytes(struct buffer_t *buff)
{
	int bytes;

	if (buff == NULL) return(0);

	if (buff->readp > buff->writep) {
		bytes = buff->readp - buff->writep;
	} else if (buff->readp < buff->writep) {
		bytes = buff->tailp - buff->writep;
	} else {
		/* (buff->readp == buff->writep) */
		if (buff->nbuffered == 0) {			/* buffer is empty */
			bytes = 0;
		} else {							/* buffer is full */
			bytes = buff->tailp - buff->writep;
		}
	}
	if (bytes <= 0) {
		if (bytes == 0) {
			;	/* too common, don't log it */
			/* debug(DBG_INF,"bfwrite() buffer \"%s\" is empty",buff->label); */
		} else {
			debug(DBG_ERR,"bfwrite() buffer \"%s\" is corrupt: bytes=%d",buff->label,bytes);
		}
	}
	return(bytes);
}

void bfdump(struct buffer_t *buff,int verbose)
{
	FILE *fp;
	int level;

	if (buff == NULL) return;

	if (verbose) {
		debug(DBG_VINF,"dumping BUFFER object \"%s\"",buff->label);
		debug(DBG_VINF,"  buffer size is %d",buff->size);
		debug(DBG_VINF,"  buffer begins at %p",buff->buffp);
		debug(DBG_VINF,"  buffer ends at %p",buff->tailp);
		debug(DBG_VINF,"  buffer read ptr is at %p",buff->readp);
		debug(DBG_VINF,"  buffer write ptr is at %p",buff->writep);
		debug(DBG_VINF,"  buffer contains %d bytes",buff->nbuffered);
	} else {
		debug(DBG_INF,"buffer \"%s\" contains %d bytes",buff->label,buff->nbuffered);
	}
	if (buff->nbuffered > 0) {
		fp = get_debugfp();
		level = get_debug_level();
		if (level >= DBG_LV5) {
			memdump((char *) buff->writep,buff->nbuffered,fp);
		}
	}
}

int bfeof(struct buffer_t *buff)
{
	if (buff == NULL) return(0);

	return(buff->eof);
}

int bfamount(struct buffer_t *buff)
{
	if (buff == NULL) return(0);

	return(buff->nbuffered);
}

unsigned char *bfpeek(struct buffer_t *buff,int *amount)
{
	if (buff == NULL) return(NULL);

	if (amount != NULL) {
		*amount = buff->nbuffered;
	}
	return(buff->writep);
}

int bfsize(struct buffer_t *buff)
{
	if (buff == NULL) return(0);

	return(buff->size);
}

unsigned char *bflabel(struct buffer_t *buff)
{
	if (buff == NULL) return((unsigned char *) "");

	return(buff->label);
}

int bfflush(int fd,struct buffer_t *buff)
{
	int ret;

	if (buff == NULL) return(0);

	while (buff->nbuffered > 0) {
		ret = bfwrite(fd,buff);
		if (ret < 0) break;
	}
	if (buff->nbuffered == 0) {
		bfinit(buff);
		return(0);
	} else {
		return(ret);
	}
}

void bfclear(struct buffer_t *buff)
{
	if (buff == NULL) return;

	bfinit(buff);
}

/*
	returns the number of characters that were appended to 
	the buffer
*/
int bfstrcat(struct buffer_t *buff,unsigned char *str)
{
	int len;

	if (buff == NULL) return(0);
	if (str == NULL) return(0);

	len = strlen((char *) str);
	if (len > (buff->size - buff->nbuffered)) {
		len = buff->size - buff->nbuffered;		/* truncate */
	}
	memcpy(buff->readp,str,len);
	bfaccount4read(buff,len);
	return(len);
}

/*
	returns the number of characters that were appended to 
	the buffer
*/
int bfstrncat(struct buffer_t *buff,unsigned char *str,int nbytes)
{
	int len;

	if (buff == NULL) return(0);
	if (str == NULL) return(0);

	len = nbytes;
	if (len > (buff->size - buff->nbuffered)) {
		len = buff->size - buff->nbuffered;		/* truncate */
	}
	memcpy(buff->readp,str,len);
	bfaccount4read(buff,len);
	return(len);
}

/*
	returns the number of characters that were appended to 
	the buffer
*/
int bfaddch(struct buffer_t *buff,unsigned char ch)
{
	if (buff == NULL) return(0);

	if ((buff->size - buff->nbuffered) > 0) {
		*(buff->readp) = ch;
		bfaccount4read(buff,1);
		return(1);
	} else {
		return(0);
	}
}

/*
	search for str in buffer
*/
unsigned char *bfstrstr(struct buffer_t *buff,unsigned char *str)
{
	unsigned char *s;

	if (buff == NULL) return(NULL);
	if (str == NULL) return(NULL);

	s = (unsigned char *) strstr((char *) buff->writep,(char *) str);
	if ((s != NULL) && (buff->writep <= s) && (s <= buff->readp)) {
		return(s);
	} else {
		return(NULL);
	}
}

/*
	search for char ch in buffer

	this version can handle embedded nulls.
*/
unsigned char *bfstrchr(struct buffer_t *buff,unsigned char ch)
{
	unsigned char *s;
	int n;

	if (buff == NULL) return(NULL);

	s = buff->writep;
	n = buff->nbuffered;
	while ((n > 0) && (*s != ch)) {
		s++;
		n--;
	}
	if (*s == ch) {
		return(s);
	} else {
		return(NULL);
	}
}

/*
	insert char ch into buffer at ptr

	returns 1 on success, 0 on failure
*/
int bfinsch(struct buffer_t *buff,unsigned char *ptr,unsigned char ch)
{
	unsigned char *s;

	if (buff == NULL) return(0);
	if (ptr == NULL) return(0);

	/* make sure ptr is within the active portion of the buffer */
	if ((buff->writep <= ptr) && (ptr <= buff->readp)) {
		/* move the chars between ptr and readp, 1 place to the right */
		s = buff->readp - 1;
		while (s >= ptr) {
			*(s+1) = *s;
			s--;
		}
		*ptr = ch;
		bfaccount4read(buff,1);
		return(1);
	} else {
		return(0);
	}
}

/*
	remove nbytes characters beginning at ptr from the buffer 

	returns the number of characters removed
*/
int bfrmstr(struct buffer_t *buff,unsigned char *ptr,int nbytes)
{
	unsigned char *s;

	if (buff == NULL) return(0);
	if (ptr == NULL) return(0);

	/* make sure ptr is within the active portion of the buffer */
	if ((buff->writep <= ptr) && (ptr <= buff->readp)) {
		if (nbytes >= buff->nbuffered) {	/* remove all chars */
			bfinit(buff);
			return(nbytes);
		}
		/* if caller asked to remove too much ... */
		if (nbytes > (buff->readp - ptr)) {
			nbytes = buff->readp - ptr;
		}
		/* remove nbytes characters between ptr and readp */
		s = ptr + nbytes;
		while (s < buff->readp) {
			*ptr++ = *s++;					/* move chars to the left */
		}
		bfaccount4read(buff,(nbytes * -1));
		return(nbytes);
	} else {
		return(0);
	}
}
