/*
	config.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	Functions which deal with the configuration file.

	Version 1.00	11/30/2001		
	Version 1.01	12/04/2001		Added LOCKDIR and LOCKTEMPLATE
	Version 1.02	12/06/2001		Added MODEM, SPEED, etc.
	Version 1.03	12/13/2001		Added FLOWCONTROL
	Version 1.04	01/14/2002		Added POLLINTERVAL
	Version 1.05	02/19/2002		Added DESCRIPTION and TRACKCARRIER
	Version 1.06	02/22/2002		Changed POLLINTERVAL to MS_POLLINTERVAL 
	                                and added LS_POLLINTERVAL.
	Version 1.07	03/06/2002		Added REPLYPURGEDATA
	Version 1.08	03/07/2002		Renamed MODEM to DEVICE
	Version 1.09	04/24/2002		Added CONNFLUSH and DISCFLUSH.  Removed 
	                                TRACKCARRIER.
	Version 1.10	10/24/2002		Added IDLETIMER.
	Version 1.11	12/03/2002		Added SENDLOGOUT.
*/

#include "telnetcpcd.h"
unsigned long def_speed = 9600L;
char *def_description = "Generic modem";

/*
	opens and parses the configuration file.  the members of 
	the config structure are filled in from the contents of 
	the config file.

	returns 0 on success, 1 on error.
*/
int read_config_file(char *file,struct config_t *conf)
{
	struct cf_entry entry;
	unsigned long token;
	int error = 0;
	int lines;
	FILE *fp;
	char linebuff[BUFSIZ];
	SERIAL_INFO *modem;
	SERIAL_INFO defaults;

	fp = fopen(file,"r");
	if (fp == NULL) {
		syslog(LOG_ERR,"cannot open config file %s",file);
		return(1);
	}

	/*
		init modem defaults
	*/
	memset(&defaults,(int) 0, sizeof(SERIAL_INFO));
	defaults.description = strdup(def_description);	/* MUST alloc memory */
	defaults.speed = def_speed;
	defaults.databits = 8;
	defaults.parity = PARITY_NONE;
	defaults.stopbits = 1;
	defaults.flowcontrol = HARDWARE_FLOW;
	defaults.conn_flush = 0;		/* don't flush serial port on connect */
	defaults.disc_flush = 1;		/* do flush serial port on discconnect */

	/*
		parse the config file and save the info within the 
		config_t structure
	*/
	lines = 0;
	modem = &defaults;					/* ptr to modem defaults */
	while (fgets(linebuff,sizeof(linebuff),fp) != NULL) {
		++lines;						/* incr config file line count */
		token = parse_entry(linebuff,&entry);
		switch (token) {
		case COMMENT:
		case BLANKLINE:
			break;						/* these are no-ops */
		case DIRECTORY:
			error = save_value(entry.value,entry.type,&(conf->directory));
			break;
		case TMPDIR:
			error = save_value(entry.value,entry.type,&(conf->tmpdir));
			break;
		case LOCKDIR:
			error = save_value(entry.value,entry.type,&(conf->lockdir));
			break;
		case LOCKTEMPLATE:
			error = save_value(entry.value,entry.type,&(conf->locktemplate));
			break;
		case DEBUGLOG:
			error = save_value(entry.value,entry.type,&(conf->debuglog));
			break;
		case DEBUGLEVEL:
			error = save_value(entry.value,entry.type,&(conf->debuglevel));
			break;
		case PIDFILE:
			error = save_value(entry.value,entry.type,&(conf->pidfile));
			break;
		case TIMEOUT:
			error = save_value(entry.value,entry.type,&(conf->timeout));
			break;
		case USERNAME:
			error = save_value(entry.value,entry.type,&(conf->user));
			break;
		case GROUP:
			error = save_value(entry.value,entry.type,&(conf->group));
			break;
		case SERVERTYPE:
			error = save_value(entry.value,entry.type,&(conf->servertype));
			break;
		case DEVICE:
			/*
				add modem to conf->modems[] array.  increments conf->nmodems 
				and dynamically allocates memory for the device name.
			*/
			error = add_modem(conf,entry.value);
			if (! error) {
				modem = conf->modems[conf->nmodems - 1];	/* ptr to current modem */
				modem->speed = defaults.speed;				/* copy defaults */
				modem->databits = defaults.databits;
				modem->parity = defaults.parity;
				modem->stopbits = defaults.stopbits;
				modem->flowcontrol = defaults.flowcontrol;
				modem->description = strdup(defaults.description);
				modem->conn_flush = defaults.conn_flush;
				modem->disc_flush = defaults.disc_flush;
			}
			break;
		case DESCRIPTION:
			if (modem->description != NULL) {
				free(modem->description);
				modem->description = NULL;
			}
			error = save_value(entry.value,entry.type,&(modem->description));
			break;
		case SPEED:
			error = save_value(entry.value,entry.type,&(modem->speed));
			if (! error) {
				if (! valid_baudrate(modem->speed)) {
					syslog(LOG_ERR,"invalid modem speed at line %d: %s",lines,entry.value);
					error = 1;
				}
			}
			break;
		case DATABITS:
			error = save_value(entry.value,entry.type,&(modem->databits));
			if (! error) {
				switch (modem->databits) {
				case 8:
				case 7:
				case 6:
				case 5:
					break;
				default:
					syslog(LOG_ERR,"invalid databits value at line %d: %s",lines,entry.value);
					error = 1;
					break;
				}
			}
			break;
		case PARITY:
			if (strcasecmp(entry.value,"none") == 0) {
				modem->parity = PARITY_NONE;
			} else if (strcasecmp(entry.value,"odd") == 0) {
				modem->parity = PARITY_ODD;
			} else if (strcasecmp(entry.value,"even") == 0) {
				modem->parity = PARITY_EVEN;
			} else {
				syslog(LOG_ERR,"invalid parity value at line %d: %s",lines,entry.value);
				error = 1;
			}
			break;
		case STOPBITS:
			error = save_value(entry.value,entry.type,&(modem->stopbits));
			if (! error) {
				switch (modem->stopbits) {
				case 1:
				case 2:
					break;
				default:
					syslog(LOG_ERR,"invalid stopbits value at line %d: %s",lines,entry.value);
					error = 1;
					break;
				}
			}
			break;
		case FLOWCONTROL:
			if (strcasecmp(entry.value,"hardware") == 0) {
				modem->flowcontrol = HARDWARE_FLOW;
			} else if (strcasecmp(entry.value,"rts/cts") == 0) {
				modem->flowcontrol = HARDWARE_FLOW;
			} else if (strcasecmp(entry.value,"software") == 0) {
				modem->flowcontrol = SOFTWARE_FLOW;
			} else if (strcasecmp(entry.value,"xon/xoff") == 0) {
				modem->flowcontrol = SOFTWARE_FLOW;
			} else if (strcasecmp(entry.value,"none") == 0) {
				modem->flowcontrol = NO_FLOWCONTROL;
			} else {
				syslog(LOG_ERR,"invalid flow control value at line %d: %s",lines,entry.value);
				error = 1;
			}
			break;
		case CONNFLUSH:
			error = save_value(entry.value,entry.type,&(modem->conn_flush));
			break;
		case DISCFLUSH:
			error = save_value(entry.value,entry.type,&(modem->disc_flush));
			break;
		case REPLYPURGEDATA:
			error = save_value(entry.value,entry.type,&(conf->reply_purge_data));
			break;
		case MS_POLLINTERVAL:
			error = save_value(entry.value,entry.type,&(conf->ms_pollinterval));
			break;
		case LS_POLLINTERVAL:
			error = save_value(entry.value,entry.type,&(conf->ls_pollinterval));
			break;
		case IDLETIMER:
			error = save_value(entry.value,entry.type,&(conf->idletimer));
			break;
		case SENDLOGOUT:
			error = save_value(entry.value,entry.type,&(conf->send_logout));
			break;
		case INVALID:
			syslog(LOG_ERR,"invalid or missing value at line %d: %s",lines,linebuff);
			/* we won't treat this as an error */
			break;
		case UNKNOWN:
		default:
			syslog(LOG_ERR,"unknown keyword at line %d: %s",lines,linebuff);
			/* we won't treat this as an error */
			break;
		}
		if (error != 0)
			break;
	} /* while */

	fclose(fp);							/* close config file */
	if (error != 0) {
		syslog(LOG_ERR,"error in configuration file near line %d",lines);
	}
	return(error);
}

/*
	skip over "white space" characters at the beginning of the 
	buffer.

	returns a pointer to the first non-space character.
*/
char *skip_white_space(char *buffer)
{
	while ((*buffer != '\0') && (isspace(*buffer)))
		++buffer;
	return(buffer);
}

/*
	build a hash table from our config file keyword table.
	returns the number of elements in the keyword table or -1 
	on error.
*/
int kw_init(struct cf_entry kw_table[])
{
	int e;								/* elements in kw table */
	int i;								/* gp int */
	ENTRY node;							/* hash table entry */

	e = 0;
	while ((kw_table[e].name != NULL) && (kw_table[e].name[0] != '\0')) {
		lowercase(kw_table[e].name);	/* fold keyword to lower case */
		++e;							/* count elements in table */
	}
	if (hcreate(e) == 0) {
		syslog(LOG_ERR,"kw_init(): unable to create hash table");
		return(-1);
	}
	for (i = 0; i < e; i++) {
		node.key = kw_table[i].name;
		node.data = (char *) &kw_table[i];
		if (hsearch(node,ENTER) == NULL) {
			syslog(LOG_ERR,"kw_init(): failed to build hash table"); 
			hdestroy();
			return(-1);
		}
	}
	return(e);
}

/*
	destroy the hash table
*/
void kw_done(void)
{
	hdestroy();
}

/*
	parse the buffer, which contains an entry from the config 
	file.  if the entry contains a valid keyword, we copy the 
	keyword into cf->name, place the keyword's token into 
	cf->token and place the keyword's token type into cf->type.
	finally, cf->value is set to point to the value of the 
	keyword, within the buffer.

	returns the keyword token, UNKNOWN, or INVALID
*/
unsigned long parse_entry(char *buffer,struct cf_entry *cf)
{
	char *p;
	char *s;
	int c;
	ENTRY *node;						/* ptr to hash table entry */
	ENTRY item;							/* hash table entry */
	struct cf_entry *entry;				/* keyword table entry */

	/* remove trailing '\n' from buffer */
	if ((s = strchr(buffer,(int) '\n')) != NULL) {
		*s = '\0';						/* replace '\n' with '\0' */
	}

	/* remove trailing '\r' from buffer */
	if ((s = strchr(buffer,(int) '\r')) != NULL) {
		*s = '\0';						/* replace '\r' with '\0' */
	}

	p = skip_white_space(buffer);
	if (strlen(p) <= 0)					/* blank line */
		return(BLANKLINE);
	if (strchr("#;",(int) *p) != NULL)	/* comment */
		return(COMMENT);
	s = p;								/* save ptr to keyword */

	while ((*p != '=') && (*p != '\0'))
		++p;							/* search for '=' */
	if (*p == '\0')
		return(UNKNOWN);
	/* p now points to '=' */
	if (s == p)							/* missing keyword */
		return(UNKNOWN);
	memset(cf->name,(int) ' ',SZ_KEYWORD);
	c = p - s;							/* number of chars */
	if (c >= SZ_KEYWORD)				/* bounds check */
		c = SZ_KEYWORD - 1;
	strncpy(cf->name,s,c);
	cf->name[c] = '\0';					/* null terminate */
	trimz(cf->name);					/* remove trailing spaces */
	lowercase(cf->name);				/* fold to lower case */
	syslog(LOG_DEBUG,"keyword \"%s\"",cf->name);

	/*
		look up this entry's keyword in the hash table to see if 
		it's valid.  if it is valid, this gives us the keyword type.
		if it is not valid, we use the STRING keyword type.
	*/
	item.key = cf->name;
	node = hsearch(item,FIND);
	if (node != NULL) {
		/* node.data can be used as a ptr to struct cf_entry */
		entry = (struct cf_entry *) node->data;
		cf->type = entry->type;			/* keyword token type */
		cf->token = entry->token;		/* keyword token */
	} else {
		cf->type = STRING;				/* keyword token type */
		cf->token = UNKNOWN;			/* keyword token */
	}
	syslog(LOG_DEBUG,"  token type: %s",tokentype2str(cf->type));
	syslog(LOG_DEBUG,"  token 0x%08x", (unsigned int) cf->token);

	/*
		parse errors from here on are due to a missing value 
		for the keyword.
	*/
	++p;								/* skip past equals sign */
	if (*p == '\0')
		return(INVALID);
	s = p;
	p = skip_white_space(s);			/* skip spaces after equals sign */
	if (*p == '\0')
		return(INVALID);
	trimz(p);							/* remove trailing spaces */
	if (*p == '\0')						/* shouldn't happen */
		return(INVALID);

	cf->value = p;						/* ptr to value within buffer */
	syslog(LOG_DEBUG,"  value \"%s\"",cf->value);
	return(cf->token);
}

char *tokentype2str(int tokentype)
{
	char *str;

	switch (tokentype) {
	case LONGVALUE:
		str = "number";
		break;
	case BOOLEAN:
		str = "boolean";
		break;
	case PRINTFSTRING:
		str = "printf string";
		break;
	case STRING:
		str = "string";
		break;
	case VALUE:
		str = "number";
		break;
	default:
		str = "unknown";
		break;
	}
	return(str);
}

/*
	fold a string to lower case
*/
void lowercase(char *str)
{
	int c;
	char *p;

	if (str == NULL) return;

	p = str;
	while (*p != '\0') {
		c = tolower((int) *p);
		*p++ = (char) c;
	}
}

/*
	save the value associated with a config file entry.
	returns 0 on success, 1 on error.
*/
int save_value(char *value,int type,void *target)
{
	int error = 0;
	char **str;
	int *i;
	unsigned long *u;

	switch (type) {
	case STRING:
	case PRINTFSTRING:
		str = (char **) target;			/* ptr cast */
		if (*str == NULL)
			*str = malloc(strlen(value) + 1);
		if (*str != NULL) {
			if (type == STRING) {
				strcpy(*str,value);		/* regular string, copy it */
			} else {					/* else, printf string */
				escstrcpy(*str,value);	/* allow character escapes */
			}
		} else {
			syslog(LOG_ERR,"cannot malloc space for string %s",value);
			error = 1;
		}
		break;
	case VALUE:
		i = (int *) target;				/* ptr cast */
		*i = atoi(value);				/* copy value */
		break;
	case BOOLEAN:
		i = (int *) target;				/* ptr cast */
		*i = booleantoi(value);			/* boolean string to integer */
		if (*i == -1) error = 1;
		break;
	case LONGVALUE:
		u = (unsigned long *) target;	/* ptr cast */
		*u = atol(value);				/* copy value */
		break;
	default:							/* unknown */
		syslog(LOG_ERR,"bad token type passed to save_value(): %d",type);
		error = 1;
		break;
	}
	return(error);
}

/*
	convert a "boolean string" value to an integer.  these strings 
	cause a return of 0:

		0 off false no

	and these cause a return of 1:

		1 on true yes

	anything else causes a return of -1
*/
int booleantoi(char *value)
{
	int ret;							/* return value */
	int i,j;							/* gp ints */
	char *str;							/* ptr to copy of value */
	char *words[2][5] = {
		{ "0", "off", "false", "no",  NULL },
		{ "1", "on",  "true",  "yes", NULL }
	};

	ret = -1;
	str = strdup(value);				/* make a copy of value */
	if (str != NULL) {
		char *p = str;
		while (*p != '\0') {
			*p = tolower(*p);			/* lower case the value */
			p++;
		}
		for (i = 0; i < 2; i++) {
			for (j = 0; words[i][j] != NULL; j++) {
				if (strcmp(str,words[i][j]) == 0) {
					ret = i;		/* use 0/1 array index as return value */
					break;
				}
			}
			if (ret != -1) break;
		}
		free(str);						/* release memory */
	}
	return(ret);
}

/*
	returns 1 if the specified baud rate is valid.
	otherwise it returns 0.
*/
int valid_baudrate(unsigned long speed)
{
	int valid;

	switch (speed) {
	case 0:
	case 50:
	case 75:
	case 110:
	case 134:
	case 150:
	case 200:
	case 300:
	case 600:
	case 1200:
	case 2400:
	case 4800:
	case 9600:
	case 19200:
	case 38400:
	case 57600:
	case 115200:
	case 230400:
	case 460800:
		valid = 1;
		break;
	default:
		valid = 0;
		break;
	}
	return(valid);
}
