/*
	debug.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	Functions to handle the debug log.

	Version 1.08	03/04/1996
	Version 1.09	08/05/1997
	Version 1.10	04/01/1998		Put pid back into log entries.
	Version 1.11	05/17/2001		varargs --> stdarg
	Version 1.12	12/13/2001		Removed heading lines, made the output
	                                lines look like syslog()'s output.  
	                                Added hostname[].
	Version 1.13	01/30/2002		Renamed the global "progname" variable 
	                                to "myprogname" to silence a compiler 
	                                warning.
	Version 1.14	02/18/2002		Outside of this source file, "debug" 
	                                and "debug_perror" are macros, if 
	                                DEBUG_TO_SYSLOG is defined.
	Version 1.15	02/20/2002		open_debug() does nothing when 
	                                DEBUG_TO_SYSLOG is defined.
	Version 1.16	02/22/2002		Removed the conditional code for 
	                                DEBUG_TO_SYSLOG.  It seems syslog() 
	                                can't handle certain characters; it 
	                                dumps core.
	Version 1.17	02/25/2002		Re-implemented the DEBUG_TO_SYSLOG code.
	                                Coupled with a fix to syslog_linedump() 
	                                in memdump.c, we should be able to send 
	                                debugging output to syslog again.
	Version 1.18	02/26/2002		The functionality of sending the debug 
	                                output to syslog() instead of a file is 
	                                now selectable at runtime by passing a 
	                                NULL or blank filename to open_debug().
	Version 1.19	03/26/2003		Use ferror() to test for write errors to
	                                the debug log and close it, if true.
	Version 1.20	08/12/2003		Use timestamp_ms() instead of timestamp()
*/

#include "telnetcpcd.h"

static char hostname[MAXHOSTNAMELEN+1];
static char *myprogname = NULL;
static FILE *debugfp = NULL;
static int debuglevel = 0;			/* disabled */

/*
	open the debug log in append mode.  

	returns the stream pointer or NULL on error.
*/
FILE *open_debug(char *file,int level,char *program)
{
	extern char hostname[];
	extern FILE *debugfp;
	extern char *myprogname;
	extern int debuglevel;
	FILE *fp;
	time_t seconds;
	char *p;

	if ((file != NULL) && (*file != '\0')) {
		fp = fopen(file,"a");
		if (fp != NULL) {
			debugfp = fp;				/* save in global variable */
		}
	} else {
		debugfp = fp = NULL;			/* direct debug output to syslog */
	}

	debuglevel = level;					/* save these in global variables */
	myprogname = program;

	/*
		get our host name and truncate it at the first '.'
	*/
	if (gethostname(hostname,MAXHOSTNAMELEN) == 0) {
		if ((p = strchr(hostname,'.')) != NULL) {
			*p = '\0';
		}
	} else {
		strcpy(hostname,"localhost");
	}

	return(fp);
}

/*
	close the debug log

	returns nothing.
*/
void close_debug(void)
{
	extern FILE *debugfp;
	extern char *myprogname;
	extern int debuglevel;

	if (debugfp != NULL) {
		fclose(debugfp);
		debugfp = NULL;
	}
	debuglevel = 0;
	myprogname = NULL;
}

int get_debug_level(void)
{
	extern int debuglevel;

	return(debuglevel);
}

void set_debug_level(int new_level)
{
	extern int debuglevel;

	if ((new_level >= DBG_LV0) && (new_level <= DBG_LV9)) {
		debuglevel = new_level;
	}
}

FILE *get_debugfp(void)
{
	extern FILE *debugfp;

	return(debugfp);
}

/*
	write our printf-style arguments to the debug log, 
	in syslog() format.

	returns nothing.
*/
void debug(int level,char *fmt, ...)
{
	extern char hostname[];
	extern FILE *debugfp;
	extern char *myprogname;
	extern int debuglevel;
	static char str[1024];
	va_list args;

	/* sanity checks */
	if (level > debuglevel) return;
	if (fmt == NULL) return;

	va_start(args,fmt);
	if (debugfp != NULL) {
		fprintf(debugfp,"%s %s %s[%d]: ",timestamp_ms(),hostname,myprogname,getpid());
		vfprintf(debugfp,fmt,args);
		fputc('\n',debugfp);
		if (! ferror(debugfp)) {
			fflush(debugfp);
		} else {
			close_debug();
		}
	} else {
		vsnprintf(str,sizeof(str),fmt,args);
		if (level == DBG_ERR) {
			syslog(LOG_ERR,"%s",str);
		} else if (level <= DBG_INF) {
			syslog(LOG_INFO,"%s",str);
		} else {
			syslog(LOG_DEBUG,"%s",str);
		}
	}
	va_end(args);
}

/*
	write the error message for the current system error to 
	the debug log
*/
void debug_perror(char *str)
{
	extern int errno;

	if (errno > 0) {
		if ((str == NULL) || (*str == '\0')) {
			debug(DBG_ERR,"%s",strerror(errno));
		} else {
			debug(DBG_ERR,"%s: %s",str,strerror(errno));
		}
	}
}
