/*
	memdump.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	Create a hex and character dump of the supplied buffer.

	This file uses 4 character tabstops.

	Version 1.01	10/23/1992
	Version 1.02	08/14/1997		Additional error checking.
	Version 1.03	02/01/2002		Added syslog_linedump().
	Version 1.04	02/25/2002		Escape '%' chars in syslog_linedump().
	Version 1.05	02/26/2002		Don't escape '%' chars in syslog_linedump().
	                                Instead, call syslog with "%s" as the format 
	                                and str as the arg list.  Using str as the 
	                                format causes a SIGSEGV in the bowels of 
	                                syslog().
	Version 1.06	03/26/2003		In linedump() flush the FILE stream after 
	                                each line.  In memdump() use ferror() to
	                                test for errors after each line.
*/

#include "telnetcpcd.h"
#define CHUNK	16						/* number of bytes we can dump */
										/* on one line */

/*
	functions in this file
*/
void linedump(char *buf,int len,FILE *stream);
void syslog_linedump(char *buf,int len);

void memdump(char *buf,int len,FILE *stream)
{
	int bytes;							/* bytes on this line */

	if (buf == NULL) return;

	for ( ; len > 0; len -= CHUNK, buf += CHUNK) {
		bytes = (len >= CHUNK) ? CHUNK : len;
		if (stream != NULL) {
			linedump(buf,bytes,stream);
			if (ferror(stream)) break;
		} else {
			syslog_linedump(buf,bytes);
		}
	}
}

void linedump(char *buf,int len,FILE *stream)
{
	int i;								/* gp int */
	char *p;							/* buf ptr */

	if ((stream == NULL) || (buf == NULL))
		return;
	if (len < 1) return;

	p = buf;							/* point to buffer */
	for (i = 0; i < len; i++,p++)		/* output hex part of dump */
		fprintf(stream,"%02x ",(unsigned char) *p);

	for (i = len+1; i <= CHUNK; i++)	/* pad w/spaces if (len < CHUNK) */
		fprintf(stream,"   ");

	fprintf(stream,"   ");				/* add spaces for separation */

	p = buf;							/* point to buffer */
	for (i = 0; i < len; i++,p++)		/* output character part of dump */
		fprintf(stream,"%c",isprint(*p) ? *p : '.');

	fprintf(stream,"\n");				/* add newline */

	if (! ferror(stream)) fflush(stream);
}

void syslog_linedump(char *buf,int len)
{
	int i;								/* gp int */
	char *p;							/* buf ptr */
	char *str;							/* the string we'll create */

	if (buf == NULL) return;
	if (len < 1) return;
	if (len > CHUNK) len = CHUNK;		/* trim */

	/* allocate memory for string */
	str = malloc((CHUNK * 4) + 16);
	if (str == NULL) return;

	/* get formatted dump */
	slinedump(str,buf,len);

	/* send it to syslog */
	syslog(LOG_DEBUG,"%s",str);			/* MUST be '(...,"%s",str)' and NOT '(...,str)' */

	/* release allocated memory */
	free(str);
}

char *slinedump(char *str,char *buf,int len)
{
	int i;								/* gp int */
	char *p;							/* buf ptr */
	char *s;							/* str ptr */

	if (str == NULL) return(NULL);
	if (buf == NULL) return(NULL);
	if (len < 1) return(NULL);
	if (len > CHUNK) len = CHUNK;		/* trim */

	s = str;							/* init str ptr */
	p = buf;							/* point to buffer */
	for (i = 0; i < len; i++,p++) {		/* output hex part of dump */
		sprintf(s,"%02x ",(unsigned char) *p);
		s += 3;
	}

	for (i = len+1; i <= CHUNK; i++) {	/* pad w/spaces if (len < CHUNK) */
		sprintf(s,"   ");
		s += 3;
	}

	sprintf(s,"   ");					/* add spaces for separation */
	s += 3;

	p = buf;							/* point to buffer */
	for (i = 0; i < len; i++,p++) {		/* output character part of dump */
		sprintf(s,"%c",isprint(*p) ? *p : '.');
		s++;
	}

	/* null terminate str */
	*s = '\0';

	return(str);
}
