/*
	modems.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	Functions which deal with SERIAL_INFO structures

	Version 1.00	12/06/2001
*/

#include "telnetcpcd.h"

/*
	add a modem to the array of SERIAL_INFO's within the config_t structure

	returns 0 on success, 1 on failure.
*/
int add_modem(struct config_t *conf,char *name)
{
	extern int errno;
	SERIAL_INFO *v;
	int i;

	/* sanity checks */
	if (conf == NULL) return(1);
	if (name == NULL) return(1);
	if (*name == '\0') return(1);

	if (conf->nmodems >= MAX_MODEMS) {
		syslog(LOG_ERR,"failed to add modem %s; already have %d modems",name,conf->nmodems);
		return(1);
	}

	v = alloc_modem();
	if (v != NULL) {
		i = conf->nmodems;
		conf->modems[i] = v;
		conf->nmodems++;
		conf->modems[i]->device = strdup(name);
		if (conf->modems[i]->device == NULL) {
			syslog(LOG_ERR,"strdup() error: %s",strerror(errno));
			return(1);
		}
	}
	return(0);
}

/*
	allocate memory for a new SERIAL_INFO

	returns a ptr to the new SERIAL_INFO or NULL on error
*/
SERIAL_INFO *alloc_modem(void)
{
	extern int errno;
	SERIAL_INFO *v;

	v = calloc(1,sizeof(SERIAL_INFO));
	if (v == NULL) 
		syslog(LOG_ERR,"alloc_modem() error: %s",strerror(errno));
	return(v);
}

/*
	free the memory used by a SERIAL_INFO

	we free() each dynamically allocated element before free()ing 
	the SERIAL_INFO itself.
*/
void free_modem(SERIAL_INFO *v)
{
	/* sanity checks */
	if (v == NULL) return;

	if (v->device != NULL)   free(v->device);
	if (v->lockfile != NULL) free(v->lockfile);
	free(v);
}

/*
	free the memory used by the array of SERIAL_INFO's
*/
void free_all_modems(SERIAL_INFO *v[])
{
	int i;

	/* sanity checks */
	if (v == NULL) return;

	for (i = 0; i < MAX_MODEMS; i++) {
		free_modem(v[i]);
	}
}

/*
	select a modem from the pool of available modems.
	returns a SERIAL_INFO ptr on success, NULL on failure.
*/
SERIAL_INFO *select_modem(struct config_t *conf)
{
	int nbusy;
	SERIAL_INFO *v;
	int r;
	char *lock;
	pid_t pid;
	time_t seconds;

	pid = getpid();					/* get our process id */
	v = NULL;						/* init return value */

	/*
		clear the busy flag on all modems
	*/
	for (r = 0; r < conf->nmodems; r++) {
		conf->modems[r]->busy = 0;
	}
	nbusy = 0;

	/*
		seed the random number generator
	*/
	seconds = time((time_t *) NULL);
	srand(seconds);
	
	/*
		try modems randomly, until we can lock one or we've tried 
		them all.
	*/
	while (nbusy < conf->nmodems) {
		/*
			generate a random integer between 0 and conf->nmodems. 
			formula borrowed from the C FAQ list (Q 13.16).
		*/
		r = (int) (rand() / ((RAND_MAX / conf->nmodems) + 1));
		syslog(LOG_DEBUG,"random modem index: %d",r);

		if ((r >= 0) && (r < conf->nmodems)) {
			/* skip this one if we've tried it already */
			if (conf->modems[r]->busy) continue;

			/* try to lock the modem */
			lock = uucp_lock(conf->modems[r]->device,pid);
			if (lock != NULL) {
				conf->modems[r]->lockfile = strdup(lock);
				v = conf->modems[r];
				break;
			} else {
				conf->modems[r]->busy = 1;
				nbusy++;
			}
		}
	}
	return(v);
}

int release_modem(SERIAL_INFO *v)
{
	int ret;

	/* sanity checks */
	if (v == NULL) return(1);
	if (v->device == NULL) return(1);

	ret = uucp_unlock(v->device);
	if (v->lockfile != NULL) {
		free(v->lockfile);
		v->lockfile = NULL;
	}
	return(ret);
}
