/*
	rfc2217.c

	Copyright (c) 2002,2003 Thomas J Pinkl <tom@pinkl.com>

	RFC2217 is the Telnet Com Port Control option.

	Version 1.00	02/20/2002
	Version 1.01	02/22/2002		Changes to modemstate logic.
	Version 1.02	03/04/2002		Misc. comments added.
	Version 1.03	03/05/2002		Added ask_client_signature flag.
	Version 1.04	03/06/2002		Added config file parameter that controls
	                                where or not we send replies to Purge Data
	                                commands.
	Version 1.05	03/07/2002		In serial_cleanup(), set the serial device 
	                                back to non-blocking mode before doing 
	                                anything else.  Fixed a bug in cs2uint() 
	                                where the wrong value was returned.  Fixed 
	                                a bug in the syslog message which reports 
	                                the current datasize in the 
	                                respond_telnet_cpc_datasize_subopt() 
	                                function.  Added telnet_cpc_datasize2str().
	Version 1.06	04/08/2002		Added tcsetattr() calls in serial_cleanup() 
	                                to make the temporary baud rate change to 
	                                B0 effective.
	Version 1.07	04/22/2002		If the client sends us a Notify Modemstate 
	                                or Linestate query, then send a reply, 
	                                without waiting for the poll interval to 
	                                expire.
	Version 1.08	04/23/2002		Changes in serial_init_termios() to reset
	                                INLCR,ICRNL,ONLCR,OCRNL.
	Version 1.09	04/24/2002		#ifdef'd out the "track carrier" code. 
	                                Added code to flush the serial port when 
	                                the client connects and disconnects.
	Version 1.10	05/02/2002		In get_modem_signals(), added conditionally
	                                compiled code (see TIOCMGET_CLOCAL_HACK) to 
	                                temporarily clear CLOCAL while we read the
	                                modem signals.  NOTE: this doesn't work; 
	                                CD is still always off on OSR5.
	Version 1.11	05/02/2002		In check_modemstate(), added conditionally 
	                                compiled code (see CD_FOLLOWS_DSR) to make
	                                the CD signal track the DSR signal.  Needed
	                                for SCO OSR5 (at least in version 5.0.5).
	Version 1.12	05/22/2002		Added the modem device name to the string 
	                                returned as the server's signature.
	Version 1.13	10/24/2002		Use snprintf() vs. sprintf().  Added the 
	                                ability to timeout idle connections, in 
	                                rfc2217_proc().
	Version 1.14	11/18/2002		Added Telnet LOGOUT option handling and 
	                                tied it into our idle timer.
	Version 1.15	11/25/2002		Commented out the Telnet LOGOUT code in 
	                                order to test C-Kermit 8.x without it.
	Version 1.16	11/25/2002		Put back the Telnet LOGOUT code. 
	Version 1.17	12/06/2002		'noquote' option
	Version 1.18	12/12/2002		Added cpc2skt buffer for internally 
	                                generated stuff that we send to the client.
	Version 1.19	03/24/2003		Added IGNCR to the list of input options 
	                                that we disable.
*/

#include "telnetcpcd.h"

/* for the tnmode[] array */
#define CLIENT 0x00
#define SERVER 0x01

/* global variables */
TELNET_OPTIONS tnoptions[MAX_TELNET_OPTIONS];
int tnmode[2];
int session_state = RESUME;
int carrier_state = NO_CARRIER;
int break_signaled = 0;
int ask_client_signature = 1;
int client_logged_in = 0;
SERIAL_INFO *si = NULL;
/* these default values are dictated by RFC2217.  don't change them! */
unsigned char linestate_mask = 0x00;
unsigned char linestate = 0;
unsigned char modemstate_mask = 0xff;
unsigned char modemstate = 0;

/*
	this function is called from handle_connection() in server.c, 
	as a result of our parent process accept()'ing a new socket 
	connection.

	returns 0 on success, non-zero on failure.
*/
int rfc2217(int sockfd)
{
	extern SERIAL_INFO *si;				/* global ptr */
	SERIAL_INFO *modem;					/* a modem from the pool */
	int fd;								/* modem fd */
	MYTERMIOS oldterm;					/* original termios */
	MYTERMIOS newterm;					/* our custom termios */
	char *p;							/* gp ptr */

	/* allocate a modem and prepare it for use */
	modem = serial_init(&fd,&oldterm,&newterm);
	if (modem == NULL) {
		p = "Unable to allocate a modem for you.\r\n";
		write(sockfd,p,strlen(p));
		return(1);
	}

	/* store a global ptr to the modem's SERIAL_INFO */
	si = modem;

	/* set socket options */
	network_init(sockfd,BLOCKING);

	/* pass data between the modem and the socket */
	rfc2217_proc(sockfd,fd,&newterm,modem);

	/* restore the modem line to its original state */
	serial_cleanup(modem,&fd,&oldterm,&newterm);
	si = NULL;

	return(0);
}

/*
	this function passes data between the modem and the socket.

	- init the telnet options structure
	- send initial telnet options: Com Port Control, Binary, etc.
	- set up select() call and enable a timeout
	- enter select() loop
	- use select() timeout to poll modem for state changes
	- also check for pending signals; SIGINT indicates BREAK condition
	- buffer data to/from modem/socket
	- handle telnet option negotiations
	- watch i/o channels for the IAC char and escape it, when present
	- exit select() loop on socket EOF

	returns 0 on success, non-zero on failure (eg. failed to allocate 
	memory for buffers).
*/
int rfc2217_proc(int sockfd,int modemfd,MYTERMIOS *newterm,SERIAL_INFO *modem)
{
	extern int errno;
	extern int signo_child;				/* what signal did we receive? */
	extern struct config_t conf;		/* built from config file */
	extern int client_logged_in;		/* is the client still "logged in"? */
	BUFFER *skt2mdm;					/* buffer for socket -> modem */
	BUFFER *mdm2skt;					/* buffer for modem -> socket */
	BUFFER *cpc2skt;					/* buffer for us -> socket */
	struct timeval tv;					/* timer for select() */
	int tvsecs;							/* what to put into tv.tv_sec */
	time_t now,elapsed,idle_cnt;
	int idle;							/* connection went idle */
	int ret;							/* return value of select() */
	int maxfd;							/* highest fd + 1 */
	int error;							/* error flag */
	fd_set orig_fds;					/* original fd struct */
	fd_set read_fds;					/* read fd struct */

	debug(DBG_INF,"sockfd fd=%d, modem fd=%d",sockfd,modemfd);

	/* allocate buffers */
	skt2mdm = bfmalloc("network",SZ_BUFFER);
	mdm2skt = bfmalloc("serial",SZ_BUFFER);
	cpc2skt = bfmalloc("options",SZ_BUFFER);
	if ((skt2mdm == NULL) || (mdm2skt == NULL) || (cpc2skt == NULL)) {
		bffree(skt2mdm);
		bffree(mdm2skt);
		bffree(cpc2skt);
		return(1);
	}
	bfdump(skt2mdm,1);					/* 1 == verbose dump */
	bfdump(mdm2skt,1);
	bfdump(cpc2skt,1);

	/* init telnet options structure and send initial options */
	telnet_init(sockfd,cpc2skt);

	/* set up select loop */
	maxfd = MAX(sockfd,modemfd) + 1;
	FD_ZERO(&orig_fds);
	FD_SET(sockfd,&orig_fds);
	FD_SET(modemfd,&orig_fds);
	error = 0;
	client_logged_in = 1;

	/* calc select() timeout value */
	idle = 0;
	idle_cnt = 0;
	tvsecs = MIN(conf.ms_pollinterval,conf.ls_pollinterval);
	if (conf.idletimer > 0) {
		tvsecs = MIN(conf.idletimer,tvsecs);
	}

	/* select loop */
	while ((client_logged_in) && (! idle) && (! error)) {

		/* check for idle connection */
		if (conf.idletimer > 0) {
			if (idle_cnt >= conf.idletimer) {
				if (! idle) {			/* we're just going into idle state */
					idle++;				/* set the idle state flag */
					debug(DBG_INF,"terminating idle connection on modem %s",modem->device);

					/* let the user know what's happening */
					bfstrcat(cpc2skt,"\r\ntelnetcpcd: terminating idle connection\r\n");
					bfwrite(sockfd,cpc2skt);

					/* tell the client to log out */
					send_telnet_option(sockfd,cpc2skt,DO,TELOPT_LOGOUT);
					msleep(250000);		/* sleep for 1/4 second */
				}

				/*
					we want to go through the select loop one last time, 
					in order to process the client's reply to DO LOGOUT.
				*/
			}
		}

		tv.tv_sec = tvsecs;				/* set timeout */
		tv.tv_usec = 0;
		read_fds = orig_fds;			/* structure copy */
		now = time((time_t *) NULL);	/* current time, in seconds */
		ret = select(maxfd,&read_fds,NULL,NULL,&tv);
		switch (ret) {
		case -1:						/* select error */
			if (errno != EINTR)
				syslog_perror("select() error");

			elapsed = time((time_t *) NULL) - now;	/* seconds spent in select() */
			idle_cnt += elapsed;		/* accumulate idle time */
			break;
		case 0:							/* select timeout */
			elapsed = time((time_t *) NULL) - now;	/* seconds spent in select() */
			idle_cnt += elapsed;		/* accumulate idle time */

			advise_client_of_state_changes(sockfd,modemfd,cpc2skt);
			break;
		default:						/* select fds for read/write */
			idle_cnt = 0;				/* reset idle counter */
			if (FD_ISSET(sockfd,&read_fds)) {
				error = read_socket(sockfd,modemfd,skt2mdm,cpc2skt);
				if (! error) {
					error = write_modem(modemfd,skt2mdm);
				}
				if (error) continue;	/* this will break the while loop */
			}
			if (FD_ISSET(modemfd,&read_fds)) {
				error = read_modem(modemfd,mdm2skt);
				if (! error) {
					error = write_socket(sockfd,mdm2skt);
				}
				if (error) continue;	/* this will break the while loop */
			}
			advise_client_of_state_changes(sockfd,modemfd,cpc2skt);
			break;
		}
		handle_pending_signal();		/* did we receive any signals? */

		/* react to loss of carrier */
#if 0	/* tjp, 04/24/2002 */
		if (modem->track_carrier) {
			if (check_carrier_state(modemfd) == LOST_CARRIER) {
				++error;				/* will break the while() loop */
				advise_client_of_state_changes(sockfd,modemfd,cpc2skt);
			}
		}
#endif
	}	/* while() loop ends */

	handle_pending_signal();			/* did we receive any signals? */

	bffree(skt2mdm);					/* release the buffers */
	bffree(mdm2skt);
	bffree(cpc2skt);
	return(0);
}

/*
	this function is called from handle_pending_signal() upon 
	receipt of a SIGINT
*/
void rfc2217_sigint(void)
{
	extern int break_signaled;
	extern unsigned char linestate;

	break_signaled = 1;
	linestate |= CPC_LINESTATE_BREAK_DETECT;
}

/*
	read from the socket 

	returns 0 on success, 1 on failure (including EOF)
*/
int read_socket(int sockfd,int modemfd,BUFFER *skt2mdm,BUFFER *cpc2skt)
{
	extern int useconds;			/* i/o wait time */
	int n;
	int ret;

	/* wait a bit before reading the socket */
	if (useconds > 0) {
		msleep(useconds);
	}

	n = bfread(sockfd,skt2mdm);
	if (n < 0) {					/* error on read */
		ret = 1;
	} else if (n == 0) {			/* no data read or EOF */
		if (bfeof(skt2mdm)) {
			ret = 1;				/* socket EOF */
			debug(DBG_INF,"read EOF on socket fd %d",sockfd);
		} else {
			ret = 0;				/* probably interrupted by a signal */
		}
	} else {						/* some data was read */
		ret = 0;
		bfdump(skt2mdm,0);			/* debugging dump of BUFFER */

		/* check for IAC char */
		if (bfstrchr(skt2mdm,IAC) != NULL) {
			process_telnet_options(sockfd,modemfd,skt2mdm,cpc2skt);
			bfdump(skt2mdm,0);		/* debugging dump of BUFFER */
		}
	}
	return(ret);
}

/*
	write to the modem

	returns 0 on success, 1 on failure
*/
int write_modem(int modemfd,BUFFER *skt2mdm)
{
	int n;
	int ret;

	n = bfwrite(modemfd,skt2mdm);
	if (n < 0) {					/* error on write */
		ret = 1;
	} else if (n == 0) {			/* no data written */
		ret = 0;
	} else {						/* some data was written */
		ret = 0;
		if (bfamount(skt2mdm) == 0)	/* no data left in buffer */
			linestate &= ~CPC_LINESTATE_DATA_READY;
		bfdump(skt2mdm,0);			/* debugging dump of BUFFER */
	}
	return(ret);
}

/*
	read from the modem 

	returns 0 on success, 1 on failure (including EOF)
*/
int read_modem(int modemfd,BUFFER *mdm2skt)
{
	extern unsigned char linestate;
	extern int noquote;				/* don't quote IAC char */
	extern int useconds;			/* i/o wait time */
	int n;
	int ret;

	/* wait a bit before reading the serial port */
	if (useconds > 0) {
		msleep(useconds);
	}

	n = bfread(modemfd,mdm2skt);
	if (n < 0) {					/* error on read */
		ret = 1;
	} else if (n == 0) {			/* no data read or EOF */
		if (bfeof(mdm2skt)) {
			ret = 1;				/* modem EOF */
			debug(DBG_INF,"read EOF on modem fd %d",modemfd);
		} else {
			ret = 0;
		}
	} else {						/* some data was read */
		ret = 0;
		linestate |= CPC_LINESTATE_DATA_READY;
		bfdump(mdm2skt,0);			/* debugging dump of BUFFER */

		/* check for IAC char unless 'noquote' option was enabled */
		if (! noquote) {
			if (bfstrchr(mdm2skt,IAC) != NULL) {
				escape_iac_chars(mdm2skt);
				bfdump(mdm2skt,0);	/* debugging dump of BUFFER */
			}
		}
	}
	return(ret);
}

/*
	write to the socket

	returns 0 on success, 1 on failure
*/
int write_socket(int sockfd,BUFFER *mdm2skt)
{
	extern int session_state;
	int n;
	int ret;

	if (session_state == SUSPEND)
		return(0);

	n = bfwrite(sockfd,mdm2skt);
	if (n < 0) {					/* error on write */
		ret = 1;
	} else if (n == 0) {			/* no data written */
		ret = 0;
	} else {						/* some data was written */
		ret = 0;
		bfdump(mdm2skt,0);			/* debugging dump of BUFFER */
	}
	return(ret);
}

/*
	this function:

	- selects a modem from the pool of available modems
    - opens the modem device file in non-blocking mode
	- saves the modem line's original termios settings
	- configures the modem line's termios settings for raw i/o
    - puts the modem device into blocking mode again
	- (optionally) flushes the modem device
	- opens a debug log, named for the selected modem device

	on success, we return a SERIAL_INFO ptr for the selected modem, 
	plus the file descriptor returned by open(), the original termios 
	settings, and the new termios settings.

	on failure, a NULL ptr is returned.
*/
SERIAL_INFO *serial_init(int *fd,MYTERMIOS *oldterm,MYTERMIOS *newterm)
{
	extern int errno;
	extern char *progname;				/* our program name */
	extern char *version;				/* version string */
	extern struct config_t conf;		/* built from config file */
	char log[PATH_MAX];					/* name of debug log */
	SERIAL_INFO *v;
	int flags;
	int ret;

	/* allocate a modem */
	v = select_modem(&conf);
	if (v == NULL) {
		syslog(LOG_ERR,"unable to allocate a modem");
		return(v);
	} else {
		syslog(LOG_INFO,"using modem %s",v->device);
	}

	/* open the modem */
	*fd = open(v->device,O_RDWR|O_NOCTTY|O_NDELAY);
	if (*fd < 0) {
		syslog(LOG_ERR,"open(%s,...) error: %s",v->device,strerror(errno));
		release_modem(v);
		return(NULL);
	}

	/* configure the termios settings */
	ret = serial_init_termios(v,fd,oldterm,newterm);
	if (ret != 0) {
		close(*fd);
		release_modem(v);
		return(NULL);
	}

	/* put the modem device back to blocking mode */
	flags = fcntl(*fd,F_GETFL,0);
	if (flags != -1) {
		flags &= ~O_NONBLOCK;
		flags = fcntl(*fd,F_SETFL,flags);
	}
	if (flags == -1) {
		syslog(LOG_ERR,"warning: cannot remove O_NONBLOCK from modem device %s (%s)",v->device,strerror(errno));
	}

	/* flush the serial port (both input and output) */
	if (v->conn_flush) {
		ret = tcflush(*fd,TCIOFLUSH);
		if (ret != 0) {
			syslog(LOG_ERR,"warning: cannot flush modem device %s (%s)",v->device,strerror(errno));
		} else {
			syslog(LOG_INFO,"flushed modem device %s",v->device);
		}
	}

	/* open the debug log */
	if ((conf.debuglog == NULL) || (*conf.debuglog == '\0') || (strcasecmp(conf.debuglog,"syslog") == 0)) {
		log[0] = '\0';				/* send debug output to syslog */
	} else {
		snprintf(log,sizeof(log),conf.debuglog,mybasename(v->device));
	}
	open_debug(log,conf.debuglevel,progname);
	debug(DBG_ERR,"%s %s",progname,version);

	/* return the SERIAL_INFO ptr */
	return(v);
}

int serial_init_termios(SERIAL_INFO *v,int *fd,MYTERMIOS *oldterm,MYTERMIOS *newterm)
{
	int ret;

	/* get two copies of the current termios settings */
	ret = tcgetattr(*fd,&(oldterm->ts));
	if (ret == 0) {
		ret = tcgetattr(*fd,&(newterm->ts));
	}
	if (ret != 0) {
		syslog(LOG_ERR,"tcgetattr() failed on modem %s",v->device);
		return(ret);
	}

	/* get two copies of the extended termiox settings */
#ifdef USE_TERMIOX
	ret = ioctl(*fd,TCGETX,&(oldterm->tx));
	if (ret == 0) {
		ret = ioctl(*fd,TCGETX,&(newterm->tx));
	}
	if (ret != 0) {
		syslog(LOG_ERR,"ioctl(fd,TCGETX,...) failed on modem %s",v->device);
		return(ret);
	}
#endif

	/* configure the modem line speed */
	ret = cfsetospeed(&(newterm->ts),ulong2speed_t(v->speed));
	if (ret == 0) {
		ret = cfsetispeed(&(newterm->ts),ulong2speed_t(v->speed));
	}
	if (ret != 0) {
		syslog(LOG_ERR,"cfsetospeed() failed on modem %s, speed %lu",v->device,v->speed);
		return(ret);
	}

	/* configure the modem line control options */
	newterm->ts.c_cflag |= (CLOCAL|CREAD);
	switch (v->parity) {
	case PARITY_ODD:
		newterm->ts.c_cflag |= PARENB;
		newterm->ts.c_cflag |= PARODD;
		newterm->ts.c_cflag &= ~CSTOPB;
		break;
	case PARITY_EVEN:
		newterm->ts.c_cflag |= PARENB;
		newterm->ts.c_cflag &= ~PARODD;
		newterm->ts.c_cflag &= ~CSTOPB;
		break;
	case PARITY_NONE:
	default:
		newterm->ts.c_cflag &= ~PARENB;
		newterm->ts.c_cflag &= ~CSTOPB;
		break;
	}
	newterm->ts.c_cflag &= ~CSIZE;
	newterm->ts.c_cflag |= uint2cs(v->databits);

	/* enable/disable hardware flow control */
	if (v->flowcontrol == HARDWARE_FLOW) {
#ifdef USE_TERMIOX
		newterm->tx.x_hflag |= (RTSXOFF|CTSXON);
#else
		newterm->ts.c_cflag |= CRTSCTS;
#endif
	} else {
#ifdef USE_TERMIOX
		newterm->tx.x_hflag &= ~(RTSXOFF|CTSXON);
#else
		newterm->ts.c_cflag &= ~CRTSCTS;
#endif
	}

	/* configure the modem line local options for raw mode */
	newterm->ts.c_lflag &= ~(ICANON|ECHO|ECHOE|ISIG);

	/* configure the modem line input options for parity */
	if (newterm->ts.c_cflag & PARENB) {
		newterm->ts.c_iflag |= (INPCK|ISTRIP);
	} else {
		newterm->ts.c_iflag &= ~(INPCK|ISTRIP);
	}

	/* send us SIGINT when a break condition is present */
	newterm->ts.c_iflag |= BRKINT;

	/* enable/disable software flow control */
	if (v->flowcontrol == SOFTWARE_FLOW) {
		newterm->ts.c_iflag |= (IXON|IXOFF|IXANY);
	} else {
		newterm->ts.c_iflag &= ~(IXON|IXOFF|IXANY);
	}

	/* disable these input options */
	newterm->ts.c_iflag &= ~(INLCR|ICRNL|IGNCR);

	/* configure the modem line output options for raw mode */
	newterm->ts.c_oflag &= ~OPOST;

	/*
		some systems (eg. Linux kernel 2.4.9) may do output processing, 
		even when OPOST has been reset.  so reset these explicitly.
	*/
	newterm->ts.c_oflag &= ~(ONLCR|OCRNL);

	/*
		since we've specified O_NDELAY on the open(), we cannot
		specify character and packet timeouts via c_cc[VMIN] 
		and c_cc[VTIME].
	*/
#if 0
	newterm->ts.c_cc[VMIN] = 0;
	newterm->ts.c_cc[VTIME] = 10;
#endif

	/* activate the new termios settings */
	ret = tcsetattr(*fd,TCSAFLUSH,&(newterm->ts));
	if (ret != 0) {
		syslog(LOG_ERR,"tcsetattr() failed on modem %s",v->device);
		tcsetattr(*fd,TCSANOW,&(oldterm->ts));
		return(ret);
	}

	/* activate the new termiox settings */
#ifdef USE_TERMIOX
	ret = ioctl(*fd,TCSETXW,&(newterm->tx));
	if (ret != 0) {
		syslog(LOG_ERR,"ioctl(fd,TCSETXW,...) failed on modem %s",v->device);
		tcsetattr(*fd,TCSANOW,&(oldterm->ts));
		return(ret);
	}
#endif

	return(0);
}

/*
	this function:

	- puts the modem device back to non-blocking mode 
	- hangs up the modem by setting the speed to B0
	- (optionally) flushes the modem device
	- restores the previous modem line termios settings 
	- closes the modem device file
	- releases the modem back to the pool
	- closes the debug log

	it returns nothing.
*/
void serial_cleanup(SERIAL_INFO *v,int *fd,MYTERMIOS *oldterm,MYTERMIOS *newterm)
{
	extern int errno;
	speed_t current_speed;
	int flags;
	int ret;

	syslog(LOG_INFO,"releasing modem %s",v->device);

	/* put the modem device back to non-blocking mode */
	flags = fcntl(*fd,F_GETFL,0);
	if (flags != -1) {
		flags |= O_NONBLOCK;
		flags = fcntl(*fd,F_SETFL,flags);
	}
	if (flags == -1) {
		syslog(LOG_ERR,"cannot set O_NONBLOCK on modem device %s (%s)",v->device,strerror(errno));
	}

	/* briefly set the speed to zero to force a hangup */
	current_speed = cfgetospeed(&(newterm->ts));
	cfsetospeed(&(newterm->ts),B0);
	tcsetattr(*fd,TCSADRAIN,&(newterm->ts));
	sleep(1);
	cfsetospeed(&(newterm->ts),current_speed);
	tcsetattr(*fd,TCSADRAIN,&(newterm->ts));

	/* flush the serial port (both input and output) */
	if (v->disc_flush) {
		ret = tcflush(*fd,TCIOFLUSH);
		if (ret != 0) {
			syslog(LOG_ERR,"warning: cannot flush modem device %s (%s)",v->device,strerror(errno));
		} else {
			syslog(LOG_INFO,"flushed modem device %s",v->device);
		}
	}

	/* restore the old termiox settings */
#ifdef USE_TERMIOX
	ioctl(*fd,TCSETXW,&(oldterm->tx));
#endif

	/* restore the old termios settings */
	tcsetattr(*fd,TCSADRAIN,&(oldterm->ts));

	/* close the modem device file */
	close(*fd);
	*fd = -1;

	/* release the modem back to the pool */
	release_modem(v);

	/* close the debug log */
	close_debug();
}

/*
	convert an unsigned long speed to its correspsonding 
	baud rate symbol
*/
speed_t ulong2speed_t(unsigned long s)
{
	speed_t speed;

	switch (s) {
	case 0:
		speed = B0;
		break;
	case 50:
		speed = B50;
		break;
	case 75:
		speed = B75;
		break;
	case 110:
		speed = B110;
		break;
	case 134:
		speed = B134;
		break;
	case 150:
		speed = B150;
		break;
	case 200:
		speed = B200;
		break;
	case 300:
		speed = B300;
		break;
	case 600:
		speed = B600;
		break;
	case 1200:
		speed = B1200;
		break;
	case 1800:
		speed = B1800;
		break;
	case 2400:
		speed = B2400;
		break;
	case 4800:
		speed = B4800;
		break;
	case 9600:
		speed = B9600;
		break;
	case 19200:
		speed = B19200;
		break;
	case 38400:
		speed = B38400;
		break;
#ifdef B57600
	case 57600:
		speed = B57600;
		break;
#endif
#ifdef B115200
	case 115200:
		speed = B115200;
		break;
#endif
#ifdef B230400
	case 230400:
		speed = B230400;
		break;
#endif
#ifdef B460800
	case 460800:
		speed = B460800;
		break;
#endif
	default:
		speed = B9600;
		break;
	}
	return(speed);
}

/*
	convert a baud rate symbol to the correspsonding 
	unsigned long speed 
	
*/
unsigned long speed_t2ulong(speed_t s)
{
	unsigned long speed;

	switch (s) {
	case B0:
		speed = 0;
		break;
	case B50:
		speed = 50;
		break;
	case B75:
		speed = 75;
		break;
	case B110:
		speed = 110;
		break;
	case B134:
		speed = 134;
		break;
	case B150:
		speed = 150;
		break;
	case B200:
		speed = 200;
		break;
	case B300:
		speed = 300;
		break;
	case B600:
		speed = 600;
		break;
	case B1200:
		speed = 1200;
		break;
	case B1800:
		speed = 1800;
		break;
	case B2400:
		speed = 2400;
		break;
	case B4800:
		speed = 4800;
		break;
	case B9600:
		speed = 9600;
		break;
	case B19200:
		speed = 19200;
		break;
	case B38400:
		speed = 38400;
		break;
#ifdef B57600
	case B57600:
		speed = 57600;
		break;
#endif
#ifdef B115200
	case B115200:
		speed = 115200;
		break;
#endif
#ifdef B230400
	case B230400:
		speed = 230400;
		break;
#endif
#ifdef B460800
	case B460800:
		speed = 460800;
		break;
#endif
	default:
		speed = 9600;
		break;
	}
	return(speed);
}

/*
	convert the number of data bits, specified as an integer, 
	to the corresponding CSn symbol.
*/
unsigned long uint2cs(unsigned int dbits)
{
	unsigned long cs;

	switch (dbits) {
	case 8:
		cs = CS8;
		break;
	case 7:
		cs = CS7;
		break;
	case 6:
		cs = CS6;
		break;
	case 5:
		cs = CS5;
		break;
	default:
		cs = CS8;
		break;
	}
	return(cs);
}

/*
	convert a data bits CSn symbol to an integer.
*/
unsigned int cs2uint(unsigned long cs)
{
	unsigned int dbits;

	switch (cs) {
	case CS8:
		dbits = 8;
		break;
	case CS7:
		dbits = 7;
		break;
	case CS6:
		dbits = 6;
		break;
	case CS5:
		dbits = 5;
		break;
	default:
		dbits = 8;
		break;
	}
	return(dbits);
}

/*
	convert a data bits CSn symbol to a string.
*/
char *cs2str(unsigned long cs)
{
	char *str;

	switch (cs) {
	case CS8:
		str = "CS8";
		break;
	case CS7:
		str = "CS7";
		break;
	case CS6:
		str = "CS6";
		break;
	case CS5:
		str = "CS5";
		break;
	default:
		str = "CS8";
		break;
	}
	return(str);
}

/*
	convert a Telnet CPC_DATASIZE_xxx symbol to a string.
*/
char *telnet_cpc_datasize2str(unsigned char datasize)
{
	char *str;

	switch (datasize) {
	case CPC_DATASIZE_CS8:
		str = "CS8";
		break;
	case CPC_DATASIZE_CS7:
		str = "CS7";
		break;
	case CPC_DATASIZE_CS6:
		str = "CS6";
		break;
	case CPC_DATASIZE_CS5:
		str = "CS5";
		break;
	default:
		str = "CS8";
		break;
	}
	return(str);
}

/*
	this function:

	- sets socket options: keepalive, non-blocking

	it returns 0 on success, non-zero on failure.
*/
int network_init(int sockfd,int blockopt)
{
	extern int errno;
	int i;
	int opt;
	int flags;
	int error;

	/*
		set the SO_KEEPALIVE socket option and set the socket 
		to non-blocking mode (ie. O_NONBLOCK).
	*/
	error = 0;
	opt = 1;
	if (setsockopt(sockfd,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(opt)) < 0) {
		syslog(LOG_ERR,"cannot set socket SO_KEEPALIVE (%s)",strerror(errno));
		error++;
	}
	flags = fcntl(sockfd,F_GETFL,0);
	if (flags != -1) {
		if (blockopt == NONBLOCKING) {
			flags |= O_NONBLOCK;
		} else {
			flags &= ~O_NONBLOCK;
		}
		flags = fcntl(sockfd,F_SETFL,flags);
	}
	if (flags == -1) {
		if (blockopt == NONBLOCKING) {
			syslog(LOG_ERR,"cannot set socket O_NONBLOCK (%s)",strerror(errno));
		} else {
			syslog(LOG_ERR,"cannot remove O_NONBLOCK from socket (%s)",strerror(errno));
		}
		error++;
	}
	return(error);
}

/*
	this function:

	- init's the telnet options structure
	- init's some global variables related to telnet options
	- sends initial telnet options: Com Port Control, Binary, etc.

	it always returns 0.
*/
int telnet_init(int sockfd,BUFFER *cpc2skt)
{
	extern TELNET_OPTIONS tnoptions[];
	extern int tnmode[];
	extern int session_state;
	extern int break_signaled;
	extern int ask_client_signature;
	int i;

	/*
		init all telnet options to "off"
	*/
	for (i = 0; i < MAX_TELNET_OPTIONS; i++) {
		tnoptions[i].sent_will = 0;
		tnoptions[i].sent_do = 0;
		tnoptions[i].sent_wont = 0;
		tnoptions[i].sent_dont = 0;
		tnoptions[i].server = 0;
		tnoptions[i].client = 0;
	}

	/* initially we're in ASCII mode */
	tnmode[CLIENT] = ASCII;
	tnmode[SERVER] = ASCII;

	/* init other global variables */
	session_state = RESUME;
	break_signaled = 0;
	ask_client_signature = 1;

	/*
		send initial telnet option negotiations
	*/
	send_telnet_option(sockfd,cpc2skt,DO,   TELOPT_COM_PORT_OPTION);
	send_telnet_option(sockfd,cpc2skt,WILL, TELOPT_BINARY);
	send_telnet_option(sockfd,cpc2skt,DO,   TELOPT_BINARY);
	send_telnet_option(sockfd,cpc2skt,WILL, TELOPT_ECHO);
	send_telnet_option(sockfd,cpc2skt,WILL, TELOPT_SGA);
	send_telnet_option(sockfd,cpc2skt,DO,   TELOPT_SGA);

	return(0);
}

/*
	send a telnet option

	returns 0 on success, 1 on failure
*/
int send_telnet_option(int sockfd,BUFFER *cpc2skt,unsigned char optcode,unsigned char option)
{
	extern int errno;
	extern int session_state;
	unsigned char optstr[4];
	int size;
	int ret;

	/* make sure we only send it once */
	if (telnet_option_was_sent(optcode,option)) {
		syslog(LOG_INFO,"telnet option %s %s already sent",telnet_optcode2str(optcode),telnet_option2str(option));
		return(0);
	}

	/* build telnet option string */
	optstr[0] = IAC;
	optstr[1] = optcode;
	optstr[2] = option;
	optstr[3] = '\0';
	size = 3;

	/* append it to the buffer */
	bfstrncat(cpc2skt,optstr,size);

	/* send it */
	if (session_state == RESUME) {
		bfdump(cpc2skt,0);		/* debugging dump of BUFFER */
		ret = bfwrite(sockfd,cpc2skt);
		if (ret >= size) {		/* buffer may have contained more than just our optstr */
			mark_telnet_option_as_sent(optcode,option);
			syslog(LOG_INFO,"sent telnet option %s %s",telnet_optcode2str(optcode),telnet_option2str(option));
			return(0);
		} else {
			syslog(LOG_ERR,"error sending telnet option %s %s",telnet_optcode2str(optcode),telnet_option2str(option));
			return(1);
		}
	}
	return(0);
}

/*
	send a telnet Com Port Control (CPC) suboption

	returns 0 on success, 1 on failure
*/
int send_telnet_cpc_suboption(int sockfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned char *command,int cmdlen)
{
	extern int errno;
	extern int session_state;
	static unsigned char optstr[MAX_TELNET_CPC_COMMAND_LEN+8];
	unsigned char *p;
	unsigned char *cp;
	unsigned long value;
	int size;
	int i;
	int ret;

	/*
		if cmdlen is 1, 2 or 4 then interpret the command as a value
	*/
	if (cmdlen == 1) {
		value = (unsigned long) *command;
	} else if (cmdlen == 2) {
		value = (unsigned long) ntohs(*((unsigned short *) command));
	} else if (cmdlen == 4) {
		value = ntohl(*((unsigned long *) command));
	} else {
		value = 0L;
	}

	/* build telnet CPC suboption string */
	optstr[0] = IAC;
	optstr[1] = SB;
	optstr[2] = TELOPT_COM_PORT_OPTION;
	optstr[3] = suboptcode;
	size = 4;
	if (cmdlen >= MAX_TELNET_CPC_COMMAND_LEN) {
		cmdlen = MAX_TELNET_CPC_COMMAND_LEN - 1;
	}
	p = &optstr[4];

	/* the command string may contain the IAC char, which must be escaped */
	cp = command;
	for (i = 0; i < cmdlen; i++) {
		if (*cp == IAC) {
			*p++ = IAC;						/* add a 2nd IAC char to optstr */
			size++;							/* adj size */
			if (size > MAX_TELNET_CPC_COMMAND_LEN) break;
		}
		*p++ = *cp++;
		size++;
		if (size > MAX_TELNET_CPC_COMMAND_LEN) break;
	}
	*p++ = IAC;
	*p++ = SE;
	*p = '\0';
	size += 2;		/* count the IAC and the SE */

	/* append it to the buffer */
	bfstrncat(cpc2skt,optstr,size);

	/* send it */
	if (session_state == RESUME) {
		bfdump(cpc2skt,0);		/* debugging dump of BUFFER */
		ret = bfwrite(sockfd,cpc2skt);
		if (ret >= size) {		/* buffer may have contained more than just our optstr */
			telnet_cpc_log_subopt("sent",suboptcode,value,command,cmdlen);
			return(0);
		} else {
			telnet_cpc_log_subopt("error sending",suboptcode,value,command,cmdlen);
			return(1);
		}
	}
	return(0);
}

/*
	returns 1 if the specified optcode was already sent for 
	the specified option.  otherwise it returns 0.
*/
int telnet_option_was_sent(unsigned char optcode,unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];
	int was_sent;

	switch (optcode) {
	case WILL:
		was_sent = tnoptions[option].sent_will;
		break;
	case DO:
		was_sent = tnoptions[option].sent_do;
		break;
	case WONT:
		was_sent = tnoptions[option].sent_wont;
		break;
	case DONT:
		was_sent = tnoptions[option].sent_dont;
		break;
	default:
		was_sent = 0;
		break;
	}
	return(was_sent);
}

/*
	mark our telnet options structure to indicate that the specified 
	opt code for the specified option, has been sent to our client.
*/
void mark_telnet_option_as_sent(unsigned char optcode,unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	switch (optcode) {
	case WILL:
		tnoptions[option].sent_will = 1;
		break;
	case DO:
		tnoptions[option].sent_do = 1;
		break;
	case WONT:
		tnoptions[option].sent_wont = 1;
		break;
	case DONT:
		tnoptions[option].sent_dont = 1;
		break;
	default:
		syslog(LOG_ERR,"unknown telnet option code: %d",(int) optcode);
		break;
	}
}

/*
	mark our telnet options structure to indicate that the specified 
	opt code for the specified option, can be sent to our client.
*/
void mark_telnet_option_as_not_sent(unsigned char optcode,unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	switch (optcode) {
	case WILL:
		tnoptions[option].sent_will = 0;
		break;
	case DO:
		tnoptions[option].sent_do = 0;
		break;
	case WONT:
		tnoptions[option].sent_wont = 0;
		break;
	case DONT:
		tnoptions[option].sent_dont = 0;
		break;
	default:
		syslog(LOG_ERR,"unknown telnet option code: %d",(int) optcode);
		break;
	}
}

/*
	mark our telnet options structure to indicate that the specified 
	option has been negotiated, server to client.
*/
void enable_telnet_server_option(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	tnoptions[option].server = 1;
}

/*
	mark our telnet options structure to indicate that the specified 
	option has not been negotiated, server to client.
*/
void disable_telnet_server_option(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	tnoptions[option].server = 0;
}

/*
	mark our telnet options structure to indicate that the specified 
	option has been negotiated, client to server.
*/
void enable_telnet_client_option(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	tnoptions[option].client = 1;
}

/*
	mark our telnet options structure to indicate that the specified 
	option has not been negotiated, client to server.
*/
void disable_telnet_client_option(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];

	tnoptions[option].client = 0;
}

/*
	is the specified telnet option enabled, server to client?
*/
int telnet_server_option_is_enabled(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];
	int enabled;

	enabled = tnoptions[option].server;
	return(enabled);
}

/*
	is the specified telnet option enabled, client to server?
*/
int telnet_client_option_is_enabled(unsigned char option)
{
	extern TELNET_OPTIONS tnoptions[];
	int enabled;

	enabled = tnoptions[option].client;
	return(enabled);
}

/*
	is the specified telnet option disabled, server to client?
*/
int telnet_server_option_is_disabled(unsigned char option)
{
	return(! telnet_server_option_is_enabled(option));
}

/*
	is the specified telnet option disabled, client to server?
*/
int telnet_client_option_is_disabled(unsigned char option)
{
	return(! telnet_client_option_is_enabled(option));
}

/*
	this function is called when the IAC char is detected in the data 
	that was read from the socket.  if the buffer contains one or more 
	telnet option strings, then we act on them and remove them from 
	the buffer (so they won't be passed on to the modem).
*/
void process_telnet_options(int sockfd,int modemfd,BUFFER *skt2mdm,BUFFER *cpc2skt)
{
	unsigned char *ptr;
	unsigned char *iac;
	unsigned char *se;
	unsigned char optcode;
	unsigned char option;
	int done;
	int size;

	/* sanity checks */
	if (skt2mdm == NULL) return;
	if (cpc2skt == NULL) return;

	/*
		peek at the buffer.  ptr points to its "active" portion 
		and size holds the number of chars buffered.
	*/
	ptr = bfpeek(skt2mdm,&size);

	/* get pointer to IAC char */
	iac = mystrchr(ptr,IAC,size);
	if (iac == NULL) return;

	debug(DBG_VINF,"processing telnet options received from client");

	/* adjust size */
	size -= (iac - ptr);
	if (size < 2) return;				/* 2 is the minimum telnet command length */

	done = 0;
	while (! done) {

		/* decide what to do next, based upon the char following the IAC */
		optcode = (unsigned char) *(iac + 1);
		switch (optcode) {
		case IAC:						/* a double IAC */
			bfrmstr(skt2mdm,iac,1);		/* remove one IAC char */
			iac++;						/* point past remaining IAC */
			size -= 2;
			break;
		case WILL:						/* WILL, DO, DONT, WONT */
		case DO:
		case DONT:
		case WONT:
			option = (unsigned char) *(iac + 2);
			respond_telnet_option(sockfd,cpc2skt,optcode,option);
			bfrmstr(skt2mdm,iac,3);		/* remove those 3 bytes */
			size -= 3;
			break;
		case SB:						/* sub-option */
			option = (unsigned char) *(iac + 2);
			if ((option == TELOPT_COM_PORT_OPTION) && (telnet_client_option_is_enabled(option))) {
				process_telnet_cpc_suboption(sockfd,modemfd,skt2mdm,cpc2skt,iac+3,size-3);
			} else {
				syslog(LOG_ERR,"ignoring telnet suboption negotiations for %s",telnet_option2str(option));
			}
			se = mystrchr(iac+3,SE,size-3);	/* find SE char */
			if (se != NULL) {
				bfrmstr(skt2mdm,iac,(se - iac + 1));	/* remove telnet options str from buffer */
				size -= (se - iac + 1);
			} else {
				iac += 3;
				size -= 3;
			}
			break;
		case GA:						/* standard telnet control functions (see RFC854) */
		case EL:
		case EC:
		case AYT:
		case AO:
		case IP:
		case BREAK:
		case DM:
		case NOP:
		default:						/* invalid char after IAC.  RFC854 says treat it as NOP. */
			syslog(LOG_ERR,"ignoring telnet %s command",telnet_optcode2str(optcode));
			bfrmstr(skt2mdm,iac,2);		/* remove those 2 bytes */
			size -= 2;
			break;
		}

		if (size < 2) {					/* 2 is the minimum telnet command length */
			++done;
		}

		/* find next IAC char */
		if (! done) {
			ptr = iac;
			iac = mystrchr(ptr,IAC,size);
			if (iac == NULL) {
				++done;
			} else {
				/* adjust size */
				size -= (iac - ptr);
				if (size < 2) {			/* 2 is the minimum telnet command length */
					++done;
				}
			}
		}
	}
}

/*
	send a response to a telnet option request

	returns 0 on success, 1 on failure
*/
int respond_telnet_option(int sockfd,BUFFER *cpc2skt,unsigned char optcode,unsigned char option)
{
	extern int errno;
	extern int client_logged_in;		/* is the client still "logged in"? */
	int ret = 0;
	unsigned char nakcode;

	syslog(LOG_INFO,"received telnet option %s %s",telnet_optcode2str(optcode),telnet_option2str(option));

	switch (option) {
	case TELOPT_COM_PORT_OPTION:
	case TELOPT_ECHO:
	case TELOPT_SGA:
		ret = respond_known_telnet_option(sockfd,cpc2skt,optcode,option);
		break;
	case TELOPT_LOGOUT:
		ret = respond_known_telnet_option(sockfd,cpc2skt,optcode,option);
		if ((optcode == WILL) || (optcode == DO)) {
			client_logged_in = 0;		/* client and server agree to the logout */
		}
		break;
	case TELOPT_BINARY:
		ret = respond_telnet_binary_option(sockfd,cpc2skt,optcode,option);
		break;
	default:
		switch (optcode) {
		case WILL:
		case WONT:
			nakcode = DONT;
			break;
		case DO:
		case DONT:
			nakcode = WONT;
			break;
		}
		/* send a negative acknowledgement */
		ret = send_telnet_option(sockfd,cpc2skt,nakcode,option);
		break;
	}

	/*
		Jeffrey Altman added similar code to sredird.c.  I'm not sure why.
	*/
#if 0
	switch (optcode) {
	case WILL:
		mark_telnet_option_as_not_sent(DO,option);
		mark_telnet_option_as_not_sent(DONT,option);
		break;
	case DO:
		mark_telnet_option_as_not_sent(WILL,option);
		mark_telnet_option_as_not_sent(WONT,option);
		break;
	case WONT:
		mark_telnet_option_as_not_sent(DO,option);
		mark_telnet_option_as_not_sent(DONT,option);
		break;
	case DONT:
		mark_telnet_option_as_not_sent(WILL,option);
		mark_telnet_option_as_not_sent(WONT,option);
		break;
	default:
		break;
	}
#endif

	return(ret);
}

/*
	respond to the optcode+option that we've received from the client
*/
int respond_known_telnet_option(int sockfd,BUFFER *cpc2skt,unsigned char optcode,unsigned char option)
{
	int ret;

	/* just in case */
	if (option == TELOPT_BINARY) {
		return(respond_telnet_binary_option(sockfd,cpc2skt,optcode,option));
	}

	ret = 0;
	switch (optcode) {
	case WILL:											/* server DO, client WILL */
		if (telnet_client_option_is_disabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,DO,option);
			enable_telnet_client_option(option);		/* server <<== client */
		}
		break;
	case WONT:											/* server DO, client WONT */
		if (telnet_client_option_is_enabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,DONT,option);
			disable_telnet_client_option(option);
		}
		break;
	case DO:											/* server WILL, client DO */
		if (telnet_server_option_is_disabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,WILL,option);
			enable_telnet_server_option(option);		/* server ==>> client */
		}
		break;
	case DONT:											/* server WILL, client DONT */
		if (telnet_server_option_is_enabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,WONT,option);
			disable_telnet_server_option(option);
		}
		break;
	default:
		break;
	}

	return(ret);
}

/*
	respond to the optcode+option that we've received from the client
*/
int respond_telnet_binary_option(int sockfd,BUFFER *cpc2skt,unsigned char optcode,unsigned char option)
{
	extern int tnmode[];
	int ret;

	/* just in case */
	if (option != TELOPT_BINARY) {
		return(respond_known_telnet_option(sockfd,cpc2skt,optcode,option));
	}

	ret = 0;
	switch (optcode) {
	case WILL:											/* server DO, client WILL */
		if (telnet_client_option_is_disabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,DO,option);
			enable_telnet_client_option(option);		/* server <<== client */
			if (tnmode[CLIENT] != BINARY) {
				syslog(LOG_INFO,"telnet connection is now in BINARY mode (server <<== client)");
				tnmode[CLIENT] = BINARY;
			}
		}
		break;
	case WONT:											/* server DO, client WONT */
		if (telnet_client_option_is_enabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,DONT,option);
			disable_telnet_client_option(option);
			if (tnmode[CLIENT] != ASCII) {
				syslog(LOG_INFO,"telnet connection is now in ASCII mode (server <<== client)");
				tnmode[CLIENT] = ASCII;
			}
		}
		break;
	case DO:											/* server WILL, client DO */
		if (telnet_server_option_is_disabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,WILL,option);
			enable_telnet_server_option(option);		/* server ==>> client */
			if (tnmode[SERVER] != BINARY) {
				syslog(LOG_INFO,"telnet connection is now in BINARY mode (server ==>> client)");
				tnmode[SERVER] = BINARY;
			}
		}
		break;
	case DONT:											/* server WILL, client DONT */
		if (telnet_server_option_is_enabled(option)) {	/* prevent option loop */
			ret = send_telnet_option(sockfd,cpc2skt,WONT,option);
			disable_telnet_server_option(option);
			if (tnmode[SERVER] != ASCII) {
				syslog(LOG_INFO,"telnet connection is now in ASCII mode (server ==>> client)");
				tnmode[SERVER] = ASCII;
			}
		}
		break;
	default:
		break;
	}
	return(ret);
}

/*
	optstr points to a Telnet CPC option string which contains at least 
	the option code, plus the IAC and SE characters.
*/
int process_telnet_cpc_suboption(int sockfd,int modemfd,BUFFER *skt2mdm,BUFFER *cpc2skt,unsigned char *optstr,int optlen)
{
	extern int session_state;
	static unsigned char command[MAX_TELNET_CPC_COMMAND_LEN];
	unsigned char suboptcode;
	unsigned char *iac;
	unsigned char *se;
	unsigned char *ptr;
	unsigned long value;
	int len;
	int n;
	int ret;

	/* sanity checking */
	if (optlen < 3) return(1);

	suboptcode = *optstr++;
	optlen--;

	/* get ptrs to IAC SE sequence which terminates the command */
	se = mystrchr(optstr,SE,optlen);	/* find SE char */
	if (se == NULL) return(1);
	iac = se - 1;						/* ptr to IAC char */
	if (*iac != IAC) return(1);

	/* calc length of command.  may be zero. */
	len = iac - optstr;
	if (len >= MAX_TELNET_CPC_COMMAND_LEN) {
		len = MAX_TELNET_CPC_COMMAND_LEN - 1;
	}

	/*
		copy command to our buffer and null-terminate it.  note that 
		the command may be empty. 
	*/
	if (len > 0) memcpy(command,optstr,len);
	command[len] = '\0';

	/*
		process IAC escapes within the command
	*/
	iac = mystrchr(command,IAC,len);
	while (iac != NULL) {
		if (*(iac+1) == IAC) {				/* double IAC */
			iac++;							/* point to 2nd IAC */
			n = len - (iac - command);		/* # chars after 2nd IAC */
			for (ptr = iac; n > 0; n--) {	/* shift chars 1 position left */
				*ptr = *(ptr+1);
				ptr++;
			}
			ptr = iac;						/* *iac is no longer IAC */
			len--;							/* adj len for removed IAC */
		} else {							/* else, not double IAC */
			ptr = iac + 1;
		}
		n = len - (ptr - command);			/* # chars left in command */
		iac = mystrchr(ptr,IAC,n);			/* any more IAC chars? */
	}

	/*
		if len is 1, 2 or 4 then interpret the command as a value
	*/
	if (len == 1) {
		value = (unsigned long) command[0];
	} else if (len == 2) {
		value = (unsigned long) ntohs(*((unsigned short *) command));
	} else if (len == 4) {
		value = ntohl(*((unsigned long *) command));
	} else {
		value = 0L;
	}

	/* log which telnet CPC suboption we've received, its args, and the arg value */
	telnet_cpc_log_subopt("received",suboptcode,value,command,len);

	/* process the telnet CPC suboption */
	ret = 0;
	switch (suboptcode) {
	case CPC_SIGNATURE_C2S:
	case CPC_SIGNATURE_S2C:
		ret = respond_telnet_cpc_signature_subopt(sockfd,cpc2skt,suboptcode,command,len);
		break;
	case CPC_SET_BAUDRATE_C2S:
	case CPC_SET_BAUDRATE_S2C:
		ret = respond_telnet_cpc_baudrate_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_SET_DATASIZE_C2S:
	case CPC_SET_DATASIZE_S2C:
		ret = respond_telnet_cpc_datasize_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_SET_PARITY_C2S:
	case CPC_SET_PARITY_S2C:
		ret = respond_telnet_cpc_parity_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_SET_STOPSIZE_C2S:
	case CPC_SET_STOPSIZE_S2C:
		ret = respond_telnet_cpc_stopsize_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_SET_CONTROL_C2S:
	case CPC_SET_CONTROL_S2C:
		ret = respond_telnet_cpc_control_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	/*
		the flow control suspend and resume commands are the only ones for 
		which no reply is expected by the client
	*/
	case CPC_FLOWCONTROL_SUSPEND_C2S:
	case CPC_FLOWCONTROL_SUSPEND_S2C:
		syslog(LOG_INFO,"telnet CPC client suspends the session");
		session_state = SUSPEND;
		break;
	case CPC_FLOWCONTROL_RESUME_C2S:
	case CPC_FLOWCONTROL_RESUME_S2C:
		syslog(LOG_INFO,"telnet CPC client resumes the session");
		session_state = RESUME;
		break;

	case CPC_SET_LINESTATE_MASK_C2S:
	case CPC_SET_LINESTATE_MASK_S2C:
	case CPC_NOTIFY_LINESTATE_C2S:
	case CPC_NOTIFY_LINESTATE_S2C:
		ret = respond_telnet_cpc_linestate_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_SET_MODEMSTATE_MASK_C2S:
	case CPC_SET_MODEMSTATE_MASK_S2C:
	case CPC_NOTIFY_MODEMSTATE_C2S:
	case CPC_NOTIFY_MODEMSTATE_S2C:
		ret = respond_telnet_cpc_modemstate_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;
	case CPC_PURGE_DATA_C2S:
	case CPC_PURGE_DATA_S2C:
		ret = respond_telnet_cpc_purge_data_subopt(sockfd,modemfd,skt2mdm,cpc2skt,suboptcode,value);
		break;
	default:
		break;
	}

	return(ret);
}

int respond_telnet_cpc_signature_subopt(int sockfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned char *command,int cmdlen)
{
	extern char *progname;				/* our program name */
	extern char *version;				/* version string */
	extern SERIAL_INFO *si;				/* global modem ptr */
	extern int ask_client_signature;	/* ask client for its signature? */
	char format[32];					/* holds format string */
	char *p;
	int len;
	int ret;

	if (*command == '\0') {			/* client wants us to send our signature */
		strcpy(format,"%s %s");		/* initial format: program name and version number */
		len = strlen(progname);
		len += strlen(version);
		len++;						/* +1 for a space between 'progname' and 'version' */
		if (si != NULL) {
			if (si->device != NULL) {
				strcat(format,", %s");
				len += 2;			/* +2 for a comma and a space */
				len += strlen(si->device);
			}
			if (si->description != NULL) {
				strcat(format,", %s");
				len += 2;			/* +2 for a comma and a space */
				len += strlen(si->description);
			}
		}
		len++;						/* +1 for a null terminator */
		p = malloc(len);
		if (p != NULL) {
			if ((si != NULL) && ((si->device != NULL) || (si->description != NULL))) {
				if ((si->device != NULL) && (si->description != NULL)) {
					snprintf(p,len,format,progname,version,si->device,si->description);
				} else {
					if (si->device != NULL) {
						snprintf(p,len,format,progname,version,si->device);
					} else {
						snprintf(p,len,format,progname,version,si->description);
					}
				}
			} else {
				snprintf(p,len,format,progname,version);
			}
			ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SIGNATURE_S2C,(unsigned char *) p,len-1);	/* -1 for the null */
			free(p);
		} else {
			syslog(LOG_ERR,"unable to allocate memory for signature");
			p = "telnetcpcd";
			ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SIGNATURE_S2C,(unsigned char *) p,strlen(p));
		}
		if (ret == 0) {				/* now request client signature, but only once */
			if (ask_client_signature) {
				ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SIGNATURE_C2S,(unsigned char *) "",0);
				ask_client_signature = 0;
			}
		}
	} else {						/* client has sent us their signature */
		syslog(LOG_INFO,"telnet CPC client signature: %s",command);
		ret = 0;
	}
	return(ret);
}

int respond_telnet_cpc_baudrate_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned int netvalue;
	int ret;

	if (value == 0L) {				/* client wants us to send the baudrate */
		value = get_baudrate(modemfd);
		netvalue = (unsigned int) htonl(value);
		syslog(LOG_INFO,"telnet CPC client requests the baudrate; sending %lu",value);
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_BAUDRATE_S2C,(unsigned char *) &netvalue,4);
	} else {						/* client wants to set a new baudrate */
		syslog(LOG_INFO,"telnet CPC client is setting the baudrate to %lu",value);
		ret = set_baudrate(modemfd,value);

		/* send new baudrate in reply */
		value = get_baudrate(modemfd);
		netvalue = (unsigned int) htonl(value);
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_BAUDRATE_S2C,(unsigned char *) &netvalue,4);
	}
	return(ret);
}

unsigned long get_baudrate(int modemfd)
{
	struct termios term;				/* current termios */
	speed_t speed;
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		speed = cfgetospeed(&term);
		return(speed_t2ulong(speed));
	} else {
		syslog_perror("tcgetattr() error");
		return(9600);
	}
}

int set_baudrate(int modemfd,unsigned long value)
{
	struct termios term;				/* current termios */
	speed_t speed;
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		speed = ulong2speed_t(value);
		cfsetospeed(&term,speed);
		ret = tcsetattr(modemfd,TCSAFLUSH,&term);
		if (ret != 0) {
			syslog_perror("tcsetattr() error");
		}
	} else {
		syslog_perror("tcgetattr() error");
	}
	return(ret);
}

int respond_telnet_cpc_datasize_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char datasize;
	int ret;

	if (value == CPC_DATASIZE_QUERY) {	/* client wants us to send the data size */
		datasize = get_datasize(modemfd);
		syslog(LOG_INFO,"telnet CPC client requests the data size; sending %s",telnet_cpc_datasize2str(datasize));
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_DATASIZE_S2C,(unsigned char *) &datasize,1);
	} else {							/* client wants to set a new data size */
		syslog(LOG_INFO,"telnet CPC client is setting the data size to %s",telnet_cpc_datasize2str((unsigned char) value));
		ret = set_datasize(modemfd,value);

		/* send new datasize in reply */
		datasize = get_datasize(modemfd);
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_DATASIZE_S2C,(unsigned char *) &datasize,1);
	}
	return(ret);
}

unsigned char get_datasize(int modemfd)
{
	struct termios term;				/* current termios */
	unsigned long cs;
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		cs = term.c_cflag & CSIZE;
		switch (cs) {
		case CS8:
			return(CPC_DATASIZE_CS8);
		case CS7:
			return(CPC_DATASIZE_CS7);
		case CS6:
			return(CPC_DATASIZE_CS6);
		case CS5:
			return(CPC_DATASIZE_CS5);
		default:
			return(CPC_DATASIZE_CS8);
		}
	} else {
		syslog_perror("tcgetattr() error");
		return(CPC_DATASIZE_CS8);
	}
}

int set_datasize(int modemfd,unsigned long value)
{
	struct termios term;				/* current termios */
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		term.c_cflag &= ~CSIZE;
		switch (value) {
		case CPC_DATASIZE_CS8:
			term.c_cflag |= CS8;
			break;
		case CPC_DATASIZE_CS7:
			term.c_cflag |= CS7;
			break;
		case CPC_DATASIZE_CS6:
			term.c_cflag |= CS6;
			break;
		case CPC_DATASIZE_CS5:
			term.c_cflag |= CS5;
			break;
		default:
			term.c_cflag |= CS8;
			break;
		}
		ret = tcsetattr(modemfd,TCSAFLUSH,&term);
		if (ret != 0) {
			syslog_perror("tcsetattr() error");
		}
	} else {
		syslog_perror("tcgetattr() error");
	}
	return(ret);
}

int respond_telnet_cpc_parity_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char parity;
	int ret;

	if (value == CPC_PARITY_QUERY) {	/* client wants us to send the parity */
		parity = get_parity(modemfd);
		syslog(LOG_INFO,"telnet CPC client requests the parity setting; sending \"%s\"",telnet_cpc_parity2str(parity));
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_PARITY_S2C,(unsigned char *) &parity,1);
	} else {							/* client wants to set a new parity setting */
		syslog(LOG_INFO,"telnet CPC client is setting the parity to \"%s\"",telnet_cpc_parity2str((unsigned char) value));
		ret = set_parity(modemfd,value);

		/* send new parity setting in reply */
		parity = get_parity(modemfd);
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_PARITY_S2C,(unsigned char *) &parity,1);
	}
	return(ret);
}

unsigned char get_parity(int modemfd)
{
	struct termios term;				/* current termios */
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		if (term.c_cflag & PARENB) {	/* if parity is enabled */
			if (term.c_cflag & PARODD) {
				return(CPC_PARITY_ODD);	/* odd parity */
			} else {
				return(CPC_PARITY_EVEN);/* even parity */
			}
		} else {
			return(CPC_PARITY_NONE);	/* no parity */
		}
	} else {
		syslog_perror("tcgetattr() error");
		return(CPC_PARITY_NONE);		/* no parity */
	}
}

int set_parity(int modemfd,unsigned long value)
{
	struct termios term;				/* current termios */
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		switch (value) {
		case CPC_PARITY_ODD:
			term.c_cflag |= PARENB;
			term.c_cflag |= PARODD;
			term.c_cflag &= ~CSTOPB;
			break;
		case CPC_PARITY_EVEN:
			term.c_cflag |= PARENB;
			term.c_cflag &= ~PARODD;
			term.c_cflag &= ~CSTOPB;
			break;
		case CPC_PARITY_MARK:
		case CPC_PARITY_SPACE:
			syslog(LOG_ERR,"ignoring request to set %s parity, will use None instead",telnet_cpc_parity2str((unsigned char) value));
			/* fall through to CPC_PARITY_NONE */
		case CPC_PARITY_NONE:
		default:
			term.c_cflag &= ~PARENB;
			term.c_cflag &= ~CSTOPB;
			break;
		}
		ret = tcsetattr(modemfd,TCSAFLUSH,&term);
		if (ret != 0) {
			syslog_perror("tcsetattr() error");
		}
	} else {
		syslog_perror("tcgetattr() error");
	}
	return(ret);
}

int respond_telnet_cpc_stopsize_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char stopbits;
	int ret;

	if (value == CPC_STOPSIZE_QUERY) {	/* client wants us to send the number of stop bits */
		stopbits = get_stopsize(modemfd);
		syslog(LOG_INFO,"telnet CPC client requests the number of stop bits; sending %s",telnet_cpc_stopsize2str(stopbits));
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_STOPSIZE_S2C,(unsigned char *) &stopbits,1);
	} else {							/* client wants to set a new number of stop bits */
		syslog(LOG_INFO,"telnet CPC client is setting the number of stop bits to %s",telnet_cpc_stopsize2str((unsigned char) value));
		ret = set_stopsize(modemfd,value);

		/* send new stopsize in reply */
		stopbits = get_stopsize(modemfd);
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_STOPSIZE_S2C,(unsigned char *) &stopbits,1);
	}
	return(ret);
}

unsigned char get_stopsize(int modemfd)
{
	struct termios term;				/* current termios */
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		if (term.c_cflag & CSTOPB) {
			return(CPC_STOPSIZE_2BITS);
		} else {
			return(CPC_STOPSIZE_1BIT);
		}
	} else {
		syslog_perror("tcgetattr() error");
		return(CPC_STOPSIZE_1BIT);
	}
}

int set_stopsize(int modemfd,unsigned long value)
{
	struct termios term;				/* current termios */
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret == 0) {
		switch (value) {
		case CPC_STOPSIZE_2BITS:
			term.c_cflag &= CSTOPB;
			break;
		case CPC_STOPSIZE_15BITS:
			syslog(LOG_ERR,"ignoring request to set 1.5 stop bits, will use 1 instead");
			/* fall through to CPC_STOPSIZE_1BIT */
		case CPC_STOPSIZE_1BIT:
		default:
			term.c_cflag &= ~CSTOPB;
			break;
		}
		ret = tcsetattr(modemfd,TCSAFLUSH,&term);
		if (ret != 0) {
			syslog_perror("tcsetattr() error");
		}
	} else {
		syslog_perror("tcgetattr() error");
	}
	return(ret);
}

int respond_telnet_cpc_control_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char flowcode;
	int ret;

	switch (value) {
	case CPC_SET_CONTROL_FLOW_QUERY:	/* client wants us to send flow control info */
	case CPC_SET_CONTROL_INFLOW_QUERY:
		flowcode = get_flowcontrol(modemfd);
		if (value == CPC_SET_CONTROL_INFLOW_QUERY) {
			/* translate CPC_SET_CONTROL_FLOW_xxx codes to CPC_SET_CONTROL_INFLOW_xxx codes */
			switch (flowcode) {
			case CPC_SET_CONTROL_FLOW_NONE:
				flowcode = CPC_SET_CONTROL_INFLOW_NONE;
				break;
			case CPC_SET_CONTROL_FLOW_XONXOFF:
				flowcode = CPC_SET_CONTROL_INFLOW_XONXOFF;
				break;
			case CPC_SET_CONTROL_FLOW_HARDWARE:
				flowcode = CPC_SET_CONTROL_INFLOW_HARDWARE;
				break;
			}
		}
		syslog(LOG_INFO,"telnet CPC client requests %sbound flow control info; sending \"%s\"",
			(value == CPC_SET_CONTROL_FLOW_QUERY) ? "out" : "in",telnet_cpc_flowcontrol2str(flowcode));
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &flowcode,1);
		break;
	case CPC_SET_CONTROL_FLOW_NONE:		/* client wants to set a new outbound flow control setting */
	case CPC_SET_CONTROL_FLOW_XONXOFF:
	case CPC_SET_CONTROL_FLOW_HARDWARE:
	case CPC_SET_CONTROL_FLOW_DCD:
	case CPC_SET_CONTROL_FLOW_DSR:
		syslog(LOG_INFO,"telnet CPC client is setting the outbound flow control to \"%s\"",
			telnet_cpc_flowcontrol2str((unsigned char) value));
		ret = set_flowcontrol(modemfd,value);

		/* send new flow control setting in reply */
		flowcode = get_flowcontrol(modemfd);
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &flowcode,1);
		break;

	/*
		since we don't support setting inbound flow control separately from 
		outbound flow control, we fake it by replying to the client with 
		whatever value was sent to us.
	*/
	case CPC_SET_CONTROL_INFLOW_NONE:	/* client wants to set a new inbound flow control setting */
	case CPC_SET_CONTROL_INFLOW_XONXOFF:
	case CPC_SET_CONTROL_INFLOW_HARDWARE:
	case CPC_SET_CONTROL_INFLOW_DTR:
		syslog(LOG_INFO,"telnet CPC client is setting the inbound flow control to \"%s\"",
			telnet_cpc_flowcontrol2str((unsigned char) value));

		/* send the requested flow control setting in reply */
		flowcode = (unsigned char) value;
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &flowcode,1);
		break;

	case CPC_SET_CONTROL_BREAK_QUERY:
	case CPC_SET_CONTROL_BREAK_ON:
	case CPC_SET_CONTROL_BREAK_OFF:
		ret = respond_telnet_cpc_control_break_subopt(sockfd,cpc2skt,suboptcode,value);
		break;

	case CPC_SET_CONTROL_DTR_QUERY:
	case CPC_SET_CONTROL_DTR_ON:
	case CPC_SET_CONTROL_DTR_OFF:
		ret = respond_telnet_cpc_control_dtr_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;

	case CPC_SET_CONTROL_RTS_QUERY:
	case CPC_SET_CONTROL_RTS_ON:
	case CPC_SET_CONTROL_RTS_OFF:
		ret = respond_telnet_cpc_control_rts_subopt(sockfd,modemfd,cpc2skt,suboptcode,value);
		break;

	default:
		ret = 0;
		break;
	}
	return(ret);
}

int respond_telnet_cpc_control_break_subopt(int sockfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	extern int break_signaled;
	unsigned char breakcode;
	int ret;

	switch (value) {
	case CPC_SET_CONTROL_BREAK_QUERY:	/* client wants us to send the break state */
		syslog(LOG_INFO,"telnet CPC client requests the break state; sending Break %s",(break_signaled) ? "On" : "Off");
		break;
	case CPC_SET_CONTROL_BREAK_ON:		/* client wants us to turn on the break state */
		syslog(LOG_INFO,"telnet CPC client sets the break state On");
		break_signaled = 1;
		break;
	case CPC_SET_CONTROL_BREAK_OFF:		/* client wants us to turn off the break state */
		syslog(LOG_INFO,"telnet CPC client sets the break state Off");
		break_signaled = 0;
		break;
	default:
		break;
	}

	/* send current break state in reply */
	if (break_signaled) {
		breakcode = CPC_SET_CONTROL_BREAK_ON;
	} else {
		breakcode = CPC_SET_CONTROL_BREAK_OFF;
	}
	ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &breakcode,1);
	return(ret);
}

int respond_telnet_cpc_control_dtr_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char dtrcode;
	unsigned int modemsignals;
	char *p;
	int ret;

	ret = 0;
	switch (value) {
	case CPC_SET_CONTROL_DTR_QUERY:		/* client wants us to send the DTR state */
		modemsignals = get_modem_signals(modemfd);
		if (modemsignals & TIOCM_DTR) {
			dtrcode = CPC_SET_CONTROL_DTR_ON;
			p = "On";
		} else {
			dtrcode = CPC_SET_CONTROL_DTR_OFF;
			p = "Off";
		}
		syslog(LOG_INFO,"telnet CPC client requests DTR state; sending DTR %s",p);
		break;
	case CPC_SET_CONTROL_DTR_ON:
		syslog(LOG_INFO,"telnet CPC client requests that DTR be raised");
		dtrcode = (unsigned char) value;
		ret = set_modem_signals(modemfd,(unsigned int) value);
		break;
	case CPC_SET_CONTROL_DTR_OFF:
		syslog(LOG_INFO,"telnet CPC client requests that DTR be lowered");
		dtrcode = (unsigned char) value;
		ret = set_modem_signals(modemfd,(unsigned int) value);
		break;
	default:
		break;
	}

	/* send current DTR value in reply */
	ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &dtrcode,1);
	return(ret);
}

int respond_telnet_cpc_control_rts_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	unsigned char rtscode;
	unsigned int modemsignals;
	char *p;
	int ret;

	ret = 0;
	switch (value) {
	case CPC_SET_CONTROL_RTS_QUERY:		/* client wants us to send the RTS state */
		modemsignals = get_modem_signals(modemfd);
		if (modemsignals & TIOCM_RTS) {
			rtscode = CPC_SET_CONTROL_RTS_ON;
			p = "On";
		} else {
			rtscode = CPC_SET_CONTROL_RTS_OFF;
			p = "Off";
		}
		syslog(LOG_INFO,"telnet CPC client requests RTS state; sending RTS %s",p);
		break;
	case CPC_SET_CONTROL_RTS_ON:
		syslog(LOG_INFO,"telnet CPC client requests that RTS be raised");
		rtscode = (unsigned char) value;
		ret = set_modem_signals(modemfd,(unsigned int) value);
		break;
	case CPC_SET_CONTROL_RTS_OFF:
		syslog(LOG_INFO,"telnet CPC client requests that RTS be lowered");
		rtscode = (unsigned char) value;
		ret = set_modem_signals(modemfd,(unsigned int) value);
		break;
	default:
		break;
	}

	/* send current RTS value in reply */
	ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_CONTROL_S2C,(unsigned char *) &rtscode,1);
	return(ret);
}

unsigned char get_flowcontrol(int modemfd)
{
	struct termios term;				/* current termios */
	int ret;
#ifdef USE_TERMIOX
	struct termiox termx;
#endif

	ret = tcgetattr(modemfd,&term);
	if (ret != 0) {
		syslog_perror("tcgetattr() error");
	}
#ifdef USE_TERMIOX
	if (ret == 0) {
		ret = ioctl(modemfd,TCGETX,&termx);
		if (ret != 0) {
			syslog_perror("ioctl(modemfd,TCGETX,) error");
		}
	}
#endif
	if (ret == 0) {
		if (term.c_iflag & IXON) {
			return(CPC_SET_CONTROL_FLOW_XONXOFF);
#ifdef USE_TERMIOX
		} else if (termx.x_hflag & (RTSXOFF|CTSXON)) {
#else
		} else if (term.c_cflag & CRTSCTS) {
#endif
			return(CPC_SET_CONTROL_FLOW_HARDWARE);
		} else {
			return(CPC_SET_CONTROL_FLOW_NONE);
		}
	} else {
		return(CPC_SET_CONTROL_FLOW_NONE);
	}
}

int set_flowcontrol(int modemfd,unsigned long value)
{
	struct termios term;				/* current termios */
#ifdef USE_TERMIOX
	struct termiox termx;
#endif
	int ret;

	ret = tcgetattr(modemfd,&term);
	if (ret != 0) {
		syslog_perror("tcgetattr() error");
	}
#ifdef USE_TERMIOX
	if (ret == 0) {
		ret = ioctl(modemfd,TCGETX,&termx);
		if (ret != 0) {
			syslog_perror("ioctl(modemfd,TCGETX,) error");
		}
	}
#endif
	if (ret == 0) {
		switch (value) {
		case CPC_SET_CONTROL_FLOW_NONE:
		case CPC_SET_CONTROL_INFLOW_NONE:
#ifdef USE_TERMIOX
			termx.x_hflag &= ~(RTSXOFF|CTSXON);
#else
			term.c_cflag &= ~CRTSCTS;
#endif
			term.c_iflag &= ~(IXON|IXOFF|IXANY);
			break;
		case CPC_SET_CONTROL_FLOW_XONXOFF:
		case CPC_SET_CONTROL_INFLOW_XONXOFF:
#ifdef USE_TERMIOX
			termx.x_hflag &= ~(RTSXOFF|CTSXON);
#else
			term.c_cflag &= ~CRTSCTS;
#endif
			term.c_iflag |= (IXON|IXOFF|IXANY);
			break;
		case CPC_SET_CONTROL_FLOW_HARDWARE:
		case CPC_SET_CONTROL_INFLOW_HARDWARE:
		case CPC_SET_CONTROL_FLOW_DCD:
		case CPC_SET_CONTROL_INFLOW_DTR:
		case CPC_SET_CONTROL_FLOW_DSR:
#ifdef USE_TERMIOX
			termx.x_hflag |= (RTSXOFF|CTSXON);
#else
			term.c_cflag |= CRTSCTS;
#endif
			term.c_iflag &= ~(IXON|IXOFF|IXANY);
			break;
		default:
			break;
		}
		ret = tcsetattr(modemfd,TCSAFLUSH,&term);
		if (ret != 0) {
			syslog_perror("tcsetattr() error");
		}
#ifdef USE_TERMIOX
		if (ret == 0) {
			ret = ioctl(modemfd,TCSETXW,&termx);
			if (ret != 0) {
				syslog_perror("ioctl(modemfd,TCSETXW,) error");
			}
		}
#endif
	}
	return(ret);
}

unsigned int get_modem_signals(int modemfd)
{
	unsigned int modemsignals;
	int ret;
#ifdef TIOCMGET_CLOCAL_HACK
	struct sigaction act;				/* new SIGHUP handler */
	struct sigaction oact;				/* saved SIGHUP handler */
	struct termios term;				/* current termios */
	int restore_term;					/* restore termios? */
	int restore_sig;					/* restore signal handler? */

	restore_term = 0;					/* initially we don't need to restore the termios settings */
	restore_sig = 0;					/*   or the SIGHUP signal handler */

	/* get the current termios settings.  if the CLOCAL bit is set ... */
	ret = tcgetattr(modemfd,&term);
	if ((ret == 0) && (term.c_cflag & CLOCAL)) {
		/* we're going to ignore SIGHUP and save the prior SIGHUP signal handler */
		act.sa_handler = SIG_IGN;
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;
		if (sigaction(SIGHUP,&act,&oact) == 0) {
			restore_sig = 1;
			/* now clear the CLOCAL bit to establish modem control */
			term.c_cflag &= ~CLOCAL;
			if (tcsetattr(modemfd,TCSANOW,&term) == 0) {
				restore_term = 1;
			} else {
				syslog(LOG_ERR,"tcsetattr() failed while trying to clear CLOCAL");
			}
		} else {
			syslog_perror("sigaction(SIGHUP,&act,&oact) error");
		}
	} else {
		if (ret != 0) {
			syslog(LOG_ERR,"tcgetattr() failed while trying to get modem signals");
		}
	}
#endif	/* TIOCMGET_CLOCAL_HACK */

	ret = ioctl(modemfd,TIOCMGET,&modemsignals);
	if (ret != 0) {
		syslog_perror("ioctl(modemfd,TIOCMGET,...) error");
		modemsignals = 0;
	}

#ifdef TIOCMGET_CLOCAL_HACK
	if (restore_term) {
		term.c_cflag |= CLOCAL;			/* set CLOCAL again */
		if (tcsetattr(modemfd,TCSANOW,&term) != 0) {
			syslog(LOG_ERR,"tcsetattr() failed while trying to set CLOCAL");
		}
	}
	if (restore_sig) {
		if (sigaction(SIGHUP,&oact,NULL) != 0) {
			syslog_perror("sigaction(SIGHUP,&oact,NULL) error");
		}
	}
#endif	/* TIOCMGET_CLOCAL_HACK */

	return(modemsignals);
}

int set_modem_signals(int modemfd,unsigned long value)
{
	unsigned int modemsignals;
	int ret;

	ret = ioctl(modemfd,TIOCMGET,&modemsignals);
	if (ret == 0) {
		switch (value) {
		case CPC_SET_CONTROL_DTR_ON:
			modemsignals |= TIOCM_DTR;
			break;
		case CPC_SET_CONTROL_DTR_OFF:
			modemsignals &= ~TIOCM_DTR;
			break;
		case CPC_SET_CONTROL_RTS_ON:
			modemsignals |= TIOCM_RTS;
			break;
		case CPC_SET_CONTROL_RTS_OFF:
			modemsignals &= ~TIOCM_RTS;
			break;
		}
		ret = ioctl(modemfd,TIOCMSET,&modemsignals);
		if (ret != 0) {
			syslog_perror("ioctl(modemfd,TIOCMSET,...) error");
		}
	} else {
		syslog_perror("ioctl(modemfd,TIOCMGET,...) error");
	}
	return(ret);
}

void advise_client_of_state_changes(int sockfd,int modemfd,BUFFER *cpc2skt)
{
	if (telnet_client_option_is_enabled(TELOPT_COM_PORT_OPTION)) {
		/* advise client of any modemstate changes */
		respond_telnet_cpc_modemstate_subopt(sockfd,modemfd,cpc2skt,CPC_NOTIFY_MODEMSTATE_S2C,0);

		/* advise client of any pending linestate changes */
		respond_telnet_cpc_linestate_subopt(sockfd,modemfd,cpc2skt,CPC_NOTIFY_LINESTATE_S2C,0);
	}
}

unsigned char check_modemstate(int modemfd,int query)
{
	extern struct config_t conf;				/* built from config file */
	extern unsigned char modemstate;
	static time_t last_ms_check = 0L;
	time_t now;
	unsigned int modemsignals;
	unsigned char new_modemstate;

	/* only update modemstate every ms_pollinterval seconds unless client sent a query */
	now = time((time_t *) NULL);
	if (! query) {
		if (now < (last_ms_check + conf.ms_pollinterval))
			return(modemstate);
	}

	modemsignals = get_modem_signals(modemfd);
	new_modemstate = 0;
	if (modemsignals & TIOCM_CAR) new_modemstate |= CPC_MODEMSTATE_CD;
	if (modemsignals & TIOCM_RNG) new_modemstate |= CPC_MODEMSTATE_RI;
	if (modemsignals & TIOCM_DSR) new_modemstate |= CPC_MODEMSTATE_DSR;
	if (modemsignals & TIOCM_CTS) new_modemstate |= CPC_MODEMSTATE_CTS;
#ifdef CD_FOLLOWS_DSR
	if (modemsignals & TIOCM_DSR) {
		new_modemstate |= CPC_MODEMSTATE_CD;
	} else {
		new_modemstate &= ~CPC_MODEMSTATE_CD;
	}
#endif
	/* calc the delta */
	if ((new_modemstate & CPC_MODEMSTATE_CD) != (modemstate & CPC_MODEMSTATE_CD)) 
		new_modemstate |= CPC_MODEMSTATE_DELTA_CD;
	if ((new_modemstate & CPC_MODEMSTATE_RI) != (modemstate & CPC_MODEMSTATE_RI)) 
		new_modemstate |= CPC_MODEMSTATE_TRLEDGE_RI;
	if ((new_modemstate & CPC_MODEMSTATE_DSR) != (modemstate & CPC_MODEMSTATE_DSR)) 
		new_modemstate |= CPC_MODEMSTATE_DELTA_DSR;
	if ((new_modemstate & CPC_MODEMSTATE_CTS) != (modemstate & CPC_MODEMSTATE_CTS)) 
		new_modemstate |= CPC_MODEMSTATE_DELTA_CTS;

	last_ms_check = now;
	return(new_modemstate);
}

int check_carrier_state(int modemfd)
{
	extern int carrier_state;
	int new_carrier_state;
	unsigned int modemsignals;

	modemsignals = get_modem_signals(modemfd);

	/* adjust carrier state */
	new_carrier_state = carrier_state;
	switch (carrier_state) {
	case NO_CARRIER:
		if (modemsignals & TIOCM_CAR) {
			syslog(LOG_INFO,"got CARRIER");
			new_carrier_state = GOT_CARRIER;
		}
		break;
	case GOT_CARRIER:
		if (! (modemsignals & TIOCM_CAR)) {
			syslog(LOG_INFO,"lost CARRIER");
			new_carrier_state = LOST_CARRIER;
		}
		break;
	case LOST_CARRIER:
		if (modemsignals & TIOCM_CAR) {
			syslog(LOG_INFO,"got CARRIER again");
			new_carrier_state = GOT_CARRIER;
		} else {
			syslog(LOG_INFO,"no CARRIER");
			new_carrier_state = NO_CARRIER;
		}
		break;
	}
	carrier_state = new_carrier_state;
	return(carrier_state);
}

int respond_telnet_cpc_linestate_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	extern struct config_t conf;				/* built from config file */
	extern unsigned char linestate_mask;
	extern unsigned char linestate;
	static time_t last_ls_check = 0L;
	time_t now;
	int query;
	int ret;
	unsigned char statechg;

	query = 0;
	ret = 0;
	switch (suboptcode) {
	case CPC_SET_LINESTATE_MASK_C2S:
	case CPC_SET_LINESTATE_MASK_S2C:
		linestate_mask = (unsigned char) value;
		syslog(LOG_INFO,"telnet CPC client sets a new linestate mask: 0x%02x",linestate_mask);

		/* send the linestate mask back in reply */
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_LINESTATE_MASK_S2C,(unsigned char *) &linestate_mask,1);
		break;
	case CPC_NOTIFY_LINESTATE_C2S:
		query = 1;						/* client sent a query */
		/* FALL THROUGH */
	case CPC_NOTIFY_LINESTATE_S2C:
		/* only update linestate every ls_pollinterval seconds (unless client sent a query) */
		now = time((time_t *) NULL);
		if ((now >= (last_ls_check + conf.ls_pollinterval)) || (query)) {
			statechg = linestate & linestate_mask;
			if (statechg) {
				syslog(LOG_INFO,"telnet CPC linestate: %s",telnet_cpc_linestate2str(statechg));
				ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_NOTIFY_LINESTATE_S2C,(unsigned char *) &statechg,1);
			}
			linestate = 0;
			last_ls_check = now;
		}
		break;
	}
	return(ret);
}

int respond_telnet_cpc_modemstate_subopt(int sockfd,int modemfd,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	extern unsigned char modemstate_mask;
	extern unsigned char modemstate;
	unsigned char new_modemstate;
	int query;
	int ret;

	query = 0;
	ret = 0;
	switch (suboptcode) {
	case CPC_SET_MODEMSTATE_MASK_C2S:
	case CPC_SET_MODEMSTATE_MASK_S2C:
		modemstate_mask = (unsigned char) value;
		syslog(LOG_INFO,"telnet CPC client sets a new modemstate mask: 0x%02x",modemstate_mask);

		/* send the modemstate mask back in reply */
		ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_SET_MODEMSTATE_MASK_S2C,(unsigned char *) &modemstate_mask,1);
		break;
	case CPC_NOTIFY_MODEMSTATE_C2S:
		query = 1;						/* client sent a query */
		/* FALL THROUGH */
	case CPC_NOTIFY_MODEMSTATE_S2C:
		new_modemstate = check_modemstate(modemfd,query);
		if ((new_modemstate != modemstate) || (query)) {	/* modemstate has changed or client sent a query */
			if (new_modemstate & modemstate_mask) {			/* is the client interested? */
				syslog(LOG_INFO,"telnet CPC modemstate: %s",telnet_cpc_modemstate2str(new_modemstate));
				ret = send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_NOTIFY_MODEMSTATE_S2C,(unsigned char *) &new_modemstate,1);
			}
			modemstate = new_modemstate;		/* update saved modemstate */
		}
		break;
	}
	return(ret);
}

int respond_telnet_cpc_purge_data_subopt(int sockfd,int modemfd,BUFFER *skt2mdm,BUFFER *cpc2skt,unsigned char suboptcode,unsigned long value)
{
	extern struct config_t conf;				/* built from config file */
	int ret;
	char *p;
	unsigned char purgecode;

	switch (value) {
	case CPC_PURGEDATA_RECVBUFF:
		ret = tcflush(modemfd,TCIFLUSH);
		bfclear(cpc2skt);
		p = "Receive buffer";
		break;
	case CPC_PURGEDATA_XMITBUFF:
		ret = tcflush(modemfd,TCOFLUSH);
		bfclear(skt2mdm);
		p = "Transmit buffer";
		break;
	case CPC_PURGEDATA_BOTH:
		ret = tcflush(modemfd,TCIOFLUSH);
		bfclear(skt2mdm);
		bfclear(cpc2skt);
		p = "both the Transmit and Receive buffers";
		break;
	default:
		ret = 0;
		p = "Unknown buffer";
		break;
	}
	syslog(LOG_INFO,"telnet CPC client requests data purge of %s",p);

	/* send the purge data command back in reply */
	if (conf.reply_purge_data) {
		purgecode = (unsigned char) value;
		ret += send_telnet_cpc_suboption(sockfd,cpc2skt,CPC_PURGE_DATA_S2C,(unsigned char *) &purgecode,1);
	}
	return(ret);
}

char *telnet_optcode2str(unsigned char optcode)
{
	char *str;

	switch (optcode) {
	case IAC:
		str = "IAC";
		break;
	case DONT:
		str = "DONT";
		break;
	case DO:
		str = "DO";
		break;
	case WONT:
		str = "WONT";
		break;
	case WILL:
		str = "WILL";
		break;
	case SB:
		str = "SB";
		break;
	case GA:
		str = "GA";
		break;
	case EL:
		str = "EL";
		break;
	case EC:
		str = "EC";
		break;
	case AYT:
		str = "AYT";
		break;
	case AO:
		str = "AO";
		break;
	case IP:
		str = "IP";
		break;
	case BREAK:
		str = "BREAK";
		break;
	case DM:
		str = "DM";
		break;
	case NOP:
		str = "NOP";
		break;
	case SE:
		str = "SE";
		break;
	default:
		str = "UNKNOWN";
		break;
	}
	return(str);
}

char *telnet_option2str(unsigned char option)
{
	static char unknown[16];
	char *str;
	char *p;

	switch (option) {
	case TELOPT_BINARY:
		str = "Binary";
		break;
	case TELOPT_ECHO:
		str = "Echo";
		break;
	case TELOPT_SGA:
		str = "Suppress Go Ahead";
		break;
	case TELOPT_LOGOUT:
		str = "Logout";
		break;
	case TELOPT_TTYPE:
		str = "Terminal Type";
		break;
	case TELOPT_NAWS:
		str = "Window Size";
		break;
	case TELOPT_XDISPLOC:
		str = "X Display Location";
		break;
	case TELOPT_AUTHENTICATION:
		str = "Authentication";
		break;
	case TELOPT_ENCRYPT:
		str = "Encryption";
		break;
	case TELOPT_NEW_ENVIRON:
		str = "New Environment";
		break;
	case TELOPT_COM_PORT_OPTION:
		str = "Com Port Control";
		break;
	case TELOPT_KERMIT:
		str = "Kermit";
		break;
	default:
		strcpy(unknown,"Option xxx");
		p = strstr(unknown,"xxx");
		snprintf(p,3,"%d",option);
		str = unknown;
		break;
	}
	return(str);
}

char *telnet_cpc_subopt2str(unsigned char suboptcode)
{
	static char unknown[16];
	char *str;
	char *p;

	switch (suboptcode) {
	case CPC_SIGNATURE_C2S:
		str = "Signature C2S";
		break;
	case CPC_SIGNATURE_S2C:
		str = "Signature S2C";
		break;
	case CPC_SET_BAUDRATE_C2S:
		str = "Set Baudrate C2S";
		break;
	case CPC_SET_BAUDRATE_S2C:
		str = "Set Baudrate S2C";
		break;
	case CPC_SET_DATASIZE_C2S:
		str = "Set Datasize C2S";
		break;
	case CPC_SET_DATASIZE_S2C:
		str = "Set Datasize S2C";
		break;
	case CPC_SET_PARITY_C2S:
		str = "Set Parity C2S";
		break;
	case CPC_SET_PARITY_S2C:
		str = "Set Parity S2C";
		break;
	case CPC_SET_STOPSIZE_C2S:
		str = "Set Stopsize C2S";
		break;
	case CPC_SET_STOPSIZE_S2C:
		str = "Set Stopsize S2C";
		break;
	case CPC_SET_CONTROL_C2S:
		str = "Set Control C2S";
		break;
	case CPC_SET_CONTROL_S2C:
		str = "Set Control S2C";
		break;
	case CPC_NOTIFY_LINESTATE_C2S:
		str = "Notify Linestate C2S";
		break;
	case CPC_NOTIFY_LINESTATE_S2C:
		str = "Notify Linestate S2C";
		break;
	case CPC_NOTIFY_MODEMSTATE_C2S:
		str = "Notify Modemstate C2S";
		break;
	case CPC_NOTIFY_MODEMSTATE_S2C:
		str = "Notify Modemstate S2C";
		break;
	case CPC_FLOWCONTROL_SUSPEND_C2S:
		str = "Suspend Flowcontrol C2S";
		break;
	case CPC_FLOWCONTROL_SUSPEND_S2C:
		str = "Suspend Flowcontrol S2C";
		break;
	case CPC_FLOWCONTROL_RESUME_C2S:
		str = "Resume Flowcontrol C2S";
		break;
	case CPC_FLOWCONTROL_RESUME_S2C:
		str = "Resume Flowcontrol S2C";
		break;
	case CPC_SET_LINESTATE_MASK_C2S:
		str = "Set Linestate Mask C2S";
		break;
	case CPC_SET_LINESTATE_MASK_S2C:
		str = "Set Linestate Mask S2C";
		break;
	case CPC_SET_MODEMSTATE_MASK_C2S:
		str = "Set Modemstate Mask C2S";
		break;
	case CPC_SET_MODEMSTATE_MASK_S2C:
		str = "Set Modemstate Mask S2C";
		break;
	case CPC_PURGE_DATA_C2S:
		str = "Purge Data C2S";
		break;
	case CPC_PURGE_DATA_S2C:
		str = "Purge Data S2C";
		break;
	default:
		strcpy(unknown,"Code xxx");
		p = strstr(unknown,"xxx");
		snprintf(p,3,"%d",suboptcode);
		str = unknown;
		break;
	}
	return(str);
}

char *telnet_cpc_parity2str(unsigned char optcode)
{
	static char unknown[16];
	char *str;
	char *p;

	switch (optcode) {
	case CPC_PARITY_NONE:
		str = "None";
		break;
	case CPC_PARITY_ODD:
		str = "Odd";
		break;
	case CPC_PARITY_EVEN:
		str = "Even";
		break;
	case CPC_PARITY_MARK:
		str = "Mark";
		break;
	case CPC_PARITY_SPACE:
		str = "Space";
		break;
	default:
		strcpy(unknown,"Code xxx");
		p = strstr(unknown,"xxx");
		snprintf(p,3,"%d",optcode);
		str = unknown;
		break;
	}
	return(str);
}

char *telnet_cpc_stopsize2str(unsigned char optcode)
{
	static char unknown[16];
	char *str;
	char *p;

	switch (optcode) {
	case CPC_STOPSIZE_1BIT:
		str = "1";
		break;
	case CPC_STOPSIZE_2BITS:
		str = "2";
		break;
	case CPC_STOPSIZE_15BITS:
		str = "1.5";
		break;
	default:
		strcpy(unknown,"Code xxx");
		p = strstr(unknown,"xxx");
		snprintf(p,3,"%d",optcode);
		str = unknown;
		break;
	}
	return(str);
}

char *telnet_cpc_modemstate2str(unsigned char modemstate)
{
	static char str[128];

	str[0] = '\0';

	if (modemstate & CPC_MODEMSTATE_CD) 
		strcat(str,"CD ");
	if (modemstate & CPC_MODEMSTATE_RI) 
		strcat(str,"RI ");
	if (modemstate & CPC_MODEMSTATE_DSR) 
		strcat(str,"DSR ");
	if (modemstate & CPC_MODEMSTATE_CTS) 
		strcat(str,"CTS ");
	if (modemstate & CPC_MODEMSTATE_DELTA_CD) 
		strcat(str,"CD~ ");
	if (modemstate & CPC_MODEMSTATE_TRLEDGE_RI) 
		strcat(str,"RI~ ");
	if (modemstate & CPC_MODEMSTATE_DELTA_DSR) 
		strcat(str,"DSR~ ");
	if (modemstate & CPC_MODEMSTATE_DELTA_CTS) 
		strcat(str,"CTS~ ");

	return(str);
}

char *telnet_cpc_linestate2str(unsigned char linestate)
{
	static char str[128];
	char *p;
	int len;

	str[0] = '\0';

	if (linestate & CPC_LINESTATE_TIMEOUT_ERROR)
		strcat(str,"Timeout Error, ");
	if (linestate & CPC_LINESTATE_TSR_EMPTY)
		strcat(str,"TSR Empty, ");
	if (linestate & CPC_LINESTATE_THR_EMPTY)
		strcat(str,"THR Empty, ");
	if (linestate & CPC_LINESTATE_BREAK_DETECT)
		strcat(str,"Break Detect, ");
	if (linestate & CPC_LINESTATE_FRAMING_ERROR)
		strcat(str,"Framing Error, ");
	if (linestate & CPC_LINESTATE_PARITY_ERROR)
		strcat(str,"Parity Error, ");
	if (linestate & CPC_LINESTATE_OVERRUN_ERROR)
		strcat(str,"Overrun Error, ");
	if (linestate & CPC_LINESTATE_DATA_READY)
		strcat(str,"Data Ready, ");

	len = strlen(str);
	p = str + len - 1;				/* point to char before the null */
									/* removing trailing ", " */
	while ((*p == ' ') || (*p == ',')) {
		p--;
	}
	p++;
	*p = '\0';
	return(str);
}

char *telnet_cpc_flowcontrol2str(unsigned char optcode)
{
	static char unknown[16];
	char *str;
	char *p;

	switch (optcode) {
	case CPC_SET_CONTROL_FLOW_NONE:
	case CPC_SET_CONTROL_INFLOW_NONE:
		str = "None";
		break;
	case CPC_SET_CONTROL_FLOW_XONXOFF:
	case CPC_SET_CONTROL_INFLOW_XONXOFF:
		str = "Xon/Xoff";
		break;
	case CPC_SET_CONTROL_FLOW_HARDWARE:
	case CPC_SET_CONTROL_INFLOW_HARDWARE:
		str = "RTS/CTS";
		break;
	default:
		strcpy(unknown,"Code xxx");
		p = strstr(unknown,"xxx");
		snprintf(p,3,"%d",optcode);
		str = unknown;
		break;
	}
	return(str);
}

/*
	this function is called when the IAC char is detected in the data 
	that was read from the modem.  for every IAC char in the buffer, 
	we "escape" it by inserting an additional IAC char.
*/
void escape_iac_chars(BUFFER *mdm2skt)
{
	unsigned char *ptr;
	unsigned char *iac;
	unsigned char *se;
	unsigned char optcode;
	unsigned char option;
	int done;
	int size;

	/* sanity checks */
	if (mdm2skt == NULL) return;

	/*
		peek at the buffer.  ptr points to its "active" portion 
		and size holds the number of chars buffered.
	*/
	ptr = bfpeek(mdm2skt,&size);

	/* get pointer to IAC char */
	iac = mystrchr(ptr,IAC,size);
	if (iac == NULL) return;

	debug(DBG_VINF,"escaping IAC chars received from modem");

	/* adjust size */
	size -= (iac - ptr);

	done = 0;
	while (! done) {
		/* insert another IAC char */
		bfinsch(mdm2skt,iac,IAC);
		size++;						/* account for added char */

		/* point past the sequence of 2 IAC chars */
		iac += 2;
		size -= 2;

		/* find next IAC char */
		if (size > 0) {
			ptr = iac;
			iac = mystrchr(ptr,IAC,size);
			if (iac == NULL) {
				++done;
			} else {
				/* adjust size */
				size -= (iac - ptr);
			}
		} else {
			++done;
		}
	}
}

/*
	log some info about a telnet CPC suboption: its name, its value, 
	and a hex dump of the command string.
*/
void telnet_cpc_log_subopt(char *prefix,unsigned char suboptcode,unsigned long value,unsigned char *command,int cmdlen)
{
	static char fmt[1024];

	fmt[0] = '\0';
	if ((prefix != NULL) && (*prefix != '\0'))
		snprintf(fmt,sizeof(fmt),"%s ",prefix);
	strcat(fmt,"telnet CPC suboption \"");
	strcat(fmt,telnet_cpc_subopt2str(suboptcode));
	strcat(fmt,"\", value ");
	if (value < 256) {
		strcat(fmt,"0x%02x");			/* one byte value, show as hex */
	} else {
		strcat(fmt,"%lu");				/* multi-byte value, show as long decimal */
	}
	if (cmdlen > 1)
		strcat(fmt,", hex dump:");

	syslog(LOG_INFO,fmt,value);			/* send to syslog */

	if (cmdlen > 1)
		memdump((char *) command,cmdlen,NULL);	/* will go to syslog() too */
}

